#!/bin/python
import os
import sys
from prody import *

from PPP.adjustments_module import WritingAtomNames, FixStructureResnames, FixAtomNames, SolveClashes
from PPP.checks_module import CheckMutation, CheckClashes
from PPP.checks_module import CheckStructure, CheckforGaps
from PPP.global_processes import ParseArguments, FindInitialAndFinalResidues, PDBwriter, RenumberStructure
from PPP.hydrogens_addition import FixStructure
from PPP.mutational_module import Mutate
from PPP.global_variables import coordination_geometries
import PPP.constraints as ct

__author__ = 'jelisa'

"""
This block adds the hid, hip, and hie residues to prody. Otherwise this would consider
this aminoacids as heteroatoms.
"""
addNonstdAminoacid('HID', 'aromatic', 'basic', 'cyclic', 'large', 'polar', 'surface')
addNonstdAminoacid('HIE', 'aromatic', 'basic', 'cyclic', 'large', 'polar', 'surface')
addNonstdAminoacid('HIP', 'aromatic', 'basic', 'cyclic', 'large', 'polar', 'surface')
addNonstdAminoacid('CYT', 'neutral', 'acyclic', 'medium', 'polar', 'buried')
addNonstdAminoacid('LYN', 'neutral', 'acyclic', 'large', 'polar', 'buried')


def main(input_pdb, pele_dir, output_pdb=["",],  no_gaps_ter=False, charge_terminals=False, make_unique=False,
         remove_terminal_missing=False, mutant_multiple=False,
         mutation="", mid_chain_nonstd_residue=[], skip=False,
         resolution=2.5, back_constr=0.5, constrain_smiles=False, ligand_pdb=False):

    if not output_pdb[0]:
        output = os.path.splitext(os.path.basename(input_pdb))[0]
        output_pdb[0] = os.path.join(pele_dir,"{}_processed.pdb".format(output))

    try:
        initial_structure = parsePDB(input_pdb)
    except IOError:
        print("The file '{}' isn't a valid file\nCheck that it does exist and try again.".format(input_pdb))
        sys.exit()
    initial_residue, final_residue = FindInitialAndFinalResidues(initial_structure)
    # ff_parameters = ReadForceFieldParameters(force_field)

    print("* Checking for insertion codes.")
    insertion_codes = [icode for icode in initial_structure.getIcodes() if icode]
    if insertion_codes:
        print("  * Due to the insertion codes the structure will be RENUMBERED starting from 1 for each chain.\n" \
              "  ** Otherwise PELE would fail.")
        structure2use = RenumberStructure(initial_structure)
    else:
        structure2use = initial_structure
    print("* Checking for gaps.")
    gaps, not_gaps = CheckforGaps(structure2use, resolution)
    if gaps is None and not_gaps is None:
        print("WARNING: Problems when checking for gaps, so don't trust the existence of gaps.")
        gaps, not_gaps = {}, {}
    print("* Checking and Fixing the Residues Names:")
    structure2use = FixStructureResnames(structure2use, make_unique)
    print("* Checking and fixing the Atoms Names:")
    structure2use = FixAtomNames(structure2use, gaps, not_gaps)
    print("* Checking the structure for missing atoms:")
    residues2fix, residues2remove, metals2coordinate, residues_without_template = CheckStructure(structure2use, gaps,
                                                                                                 not_gaps,
                                                                                                 charge_terminals,
                                                                                                 remove_terminal_missing)
    if residues2fix:
        print('* Placing the missing atoms and removing the extra atoms:')
        structure2use = FixStructure(structure2use, residues2fix, gaps, charge_terminals)

    if not mutation:
        print('Writing the structure to {}'.format(output_pdb[0]))
        if make_unique:
            ligand_chain = structure2use.select("chain {}".format(make_unique))
            if ligand_chain:
                not_proteic_ligand = structure2use.select("chain {}".format(make_unique)).hetero
            else:
                not_proteic_ligand = None
            PDBwriter(output_pdb[0], WritingAtomNames(structure2use), make_unique, residues2remove,
                      no_gaps_ter, not_proteic_ligand, gaps, not_gaps)
        else:
            not_proteic_ligand = None
            if not skip:
                PDBwriter(output_pdb[0], WritingAtomNames(structure2use), make_unique, residues2remove,
                      no_gaps_ter, not_proteic_ligand, gaps, not_gaps, mid_chain_nonstd_residue)
            else:
                shutil.copy(input_pdb, output_pdb[0])

        ########METAL ATOMS PROCESS###############
        coordinated_atoms_ids = {}
        for metal, atoms_list in metals2coordinate.items():
            metal_id = "{} {} {}".format(WritingAtomNames(metal).getNames()[0].replace(' ','_'),
                                         metal.getChid(),
                                         metal.getResnum())
            atoms_ids = [["{} {} {} {}".format(at.getResname(),
                                               at.getResnum(), at.getChid(),
                                               WritingAtomNames(at).getNames()[0].replace(' ', '_'),),
                          calcDistance(metal, at)[0]] for at in atoms_list]
            if len(atoms_list) in [x[1] for x in coordination_geometries.itervalues()]:
                coordinated_atoms_ids[metal_id] = atoms_ids

        ###########RETRIEVE CONSTANTS###############
        constr = ct.retrieve_constraints(output_pdb[0], ligand_pdb, gaps, coordinated_atoms_ids, back_constr=back_constr,
            constrain_smiles=constrain_smiles)



        return output_pdb[0], residues_without_template, gaps, coordinated_atoms_ids, constr

if __name__ == '__main__':
    arguments = ParseArguments()
    if arguments is None:
        sys.exit()
    else:
        main(arguments.input_pdb, arguments.pdb_resolution, output_pdb=arguments.output_pdb,
             no_gaps_ter=arguments.no_gaps_ter, charge_terminals=arguments.charge_terminals,
             make_unique=arguments.make_unique, remove_terminal_missing=arguments.remove_terminal_missing,
mutant_multiple=arguments.mutant_multiple, mutation=arguments.mutation)



import sys

from prody import Contacts, calcDistance, calcAngle

from PPP.global_processes import FindInitialAndFinalResidues
from PPP.global_variables import supported_aminoacids, supported_metals, coordination_geometries
from PPP.program_own_classes import ZMATRIX

__author__ = 'jelisa'

fixed_atoms = ["N", "H", "CA", "HA", "C", "O", "CB"]


def CheckCysteines(structure):
    """
    This function gets one structure and finds the cysteines that form disulfide-bridges and
     the ones that should be charged and returns their residue numbers and chain.

    :param structure: Prody Atomgroup object
    :return: two different lists
    """
    crosslinked_cys = []
    charged_cys = []
    if structure.select('resname CYS') is not None:
        for cysteine in structure.select('resname CYS').copy().iterResidues():
            if "HG" in cysteine.getNames():
                continue
            key = "{}_{}".format(cysteine.getChid(), cysteine.getResnum())
            if key in crosslinked_cys:
                continue
            elif key in charged_cys:
                print("CYS/CYT classification WTF.0??\n cystein {} {}".format(cysteine.getResnum(),
                                                                              cysteine.getChid()))
                continue
            sg = cysteine.select('name SG')
            if sg is None:
                continue
            selection_string = "(within 2.1 of atom) and (not (resnum {} and name SG))".format(cysteine.getResnum())
            neighbours = structure.select(selection_string, atom=sg)
            if neighbours is None:
                print("CYS/CYT classification WTF.1??")
            elif "SG" in neighbours.getNames():
                crosslinked_cys.append(key)
                key2 = "{}_{}".format(neighbours.select('name SG').getChids()[0],
                                      neighbours.select('name SG').getResnums()[0])
                crosslinked_cys.append(key2)

            else:
                charged_cys.append(key)
    if crosslinked_cys:
        writing_list = ["     * CYS_{} - CYS_{}".format(x, y) for x,y in zip(crosslinked_cys[0::2], crosslinked_cys[1::2])]
        print("  * The following cysteins will be considered as cross-linked:\n{0}".format("\n".join(writing_list)))
    if charged_cys:
        writing_list = ["     * CYS_{}".format(k) for k in charged_cys]
        print("  * The following cysteins will be considered as cross-linked:\n{0}".format("\n".join(writing_list)))
    return crosslinked_cys, charged_cys


# def CheckTetrahedricConformation(angles):
#     right_conformation = False
#     distorted_angles = []
#     ok_angles = []
#     accepted_angles = coordination_geometries['tetrahedric']
#     for angle in angles:
#         if angle[3] < accepted_angles[0] - 15 or angle[3] > accepted_angles[0] + 15:
#             distorted_angles.append(angle)
#         else:
#             ok_angles.append(angle)
#     if len(ok_angles) > len(distorted_angles):
#         if distorted_angles:
#             print "       * The following angle(s) are distorted for this geometry:\n{}".format(
#                 "       * {} {} {}".format(angle[0].getName(), angle[0].getResnum(), angle[0].getChid()))
#         right_conformation = True
#     return  right_conformation


def CheckConformation(angles, conformation):
    """
    A function to check that the angles are within the right range for a given conformation
    :param angles: a list containing the three atoms involved in the angle (prody instances) and the
    angle in the position 3
    :param conformation: a string containing the name of the conformation to check, it should be one defined in
    global_variables.coordination_geometries
    :return: a boolean indicating whether the conformation is right or not
    """
    right_conformation = False
    distorted_angles = []
    ok_angles = []
    accepted_angles, needed_atoms = coordination_geometries[conformation]
    if conformation not in coordination_geometries.keys():
        print("      * The desired coordination isn't implemented.")
    else:
        if len(angles) < needed_atoms:
            print("       * WARNING: The metal is missing atoms to coordinated with")

    for angle in angles:
        for a_angle in accepted_angles:
            if a_angle - 15 <= angle[3] <= a_angle + 15:
                ok_angles.append(angle)
            else:
                distorted_angles.append(angle)
    if len(ok_angles) > len(distorted_angles):
        if distorted_angles and len(angles) != needed_atoms:
            print("       * The angles match the {} coordination".format(conformation))
            # print [[[x.getName(), x.getResnum(), x.getChid()] for x in angle[]] for angle in distorted_angles]
            print("       * The angle(s) formed by the following atoms are distorted for this geometry:")#\n{}".format(
            for angle in distorted_angles:
                print("{} {} {} {}".format("        * ",
                                           "{}-{}-{}".format(angle[0].getName(), angle[0].getResnum(), angle[0].getChid(),),
                                           "{}-{}-{}".format(angle[1].getNames()[0], angle[1].getResnums()[0], angle[1].getChid()),
                                           "{}-{}-{}".format(angle[2].getName(), angle[2].getResnum(), angle[2].getChid())))
        else:
            print("       * The angles match perfectly the {} coordination".format(conformation))
        right_conformation = True
    return right_conformation


def CheckMetalsCoordination(structure):
    """
    A function to detect the metal atoms that could be coordinated with the protein.
    :param structure:
    :return:
    """
    selection_pattern = "(within 3 of metal) and (not resnum {}) and (not hydrogen) and (not carbon)"
    coordinated_metals = {}
    for metal in supported_metals:
        if structure.select('resname {}'.format(metal)) is not None:
            print("  * Checking the metals that can be coordinated. (" \
                  "a constraint should be used if they're really coordinated)")
            for metal_res in structure.select('resname {}'.format(metal)).copy().iterResidues():
                coordinated_atoms_list = []
                if metal_res.numAtoms() != 1:
                    continue
                coordinated_atoms = structure.select(selection_pattern.format(metal_res.getResnum()),
                                                     metal=metal_res)
                if coordinated_atoms is None:
                    print("     * The metal atom {} isn't coordinated with the protein. Are you sure it's necessary?")
                else:
                    prev_atom = None
                    for at in coordinated_atoms.iterAtoms():
                        if prev_atom is None:
                            coordinated_atoms_list.append(at)
                            prev_atom = at
                        else:
                            if at.getResnum() == prev_atom.getResnum() and \
                                    (at.getIndex() == prev_atom.getIndex() + 1 or at.getResname() in ['ASP', 'GLU']):
                                if calcDistance(at, metal_res) > calcDistance(prev_atom, metal_res):
                                    prev_atom = None
                                    continue
                                else:
                                    coordinated_atoms_list.pop(-1)
                            coordinated_atoms_list.append(at)
                            prev_atom = at
                    coordinated_metals[metal_res] = coordinated_atoms_list#, 'angles': angles}
    coordinated_atoms_ids = {}
    if coordinated_metals:
        for metal, atoms_list in coordinated_metals.iteritems():
            metal_id = "{} {} {}".format(metal.getResname(), metal.getChid(), metal.getResnum())
            atoms_ids = [["{} {} {} {}".format(at.getResnum(), at.getResname(), at.getChid(), at.getName()),
                          calcDistance(metal, at)[0]] for at in atoms_list]
            if len(atoms_list) in [x[1] for x in coordination_geometries.itervalues()]:
                coordinated_atoms_ids[metal_id] = atoms_ids
            print("     * The metal atom {0} has the following atoms within coordination " \
                  "distance:\n{1}".format(metal_id, "\n".join(['       * {0}'.format(x[0]) for x in atoms_ids])))
            angles = [[at, metal, at2, calcAngle(at, metal, at2)[0]]
                      for idx,at in enumerate(atoms_list) for at2 in atoms_list[idx + 1:]]
            if len(atoms_list) <= 4:
                print("      * Checking for a tetrahedric coordination for the atom.")
                found_conformation = CheckConformation(angles, 'tetrahedric')
                if not found_conformation:
                    print("        * WARNING: The angles are too distorted to ascertain this configuration. CHECK IT manually")
            elif 4 < len(atoms_list) <= 6: #or not found_conformation:
                print("      * WARNING: Checking for an octahedric coordination for the atom.")
                found_conformation = CheckConformation(angles, 'octahedric')
                if not found_conformation:
                    print("        * The angles are too distorted to ascertain this configuration. CHECK IT manually")
            else:
                print("      * The metal doesn't have a coordination we can validate.")


    else:
        print("  * There are no coordinated metals.")

    return coordinated_metals


def CheckStructure(initial_structure, gaps={}, no_gaps={}, charge_terminals=False, remove_missing_ter=False,
                   debug=False):
    residues2fix = {}
    crosslinked_cysteines, charged_cysteines = CheckCysteines(initial_structure)
    residues2remove = {}
    metals2coordinate = {}
    #metals2coordinate = CheckMetalsCoordination(initial_structure)
    residues_without_template = []
    for chain in initial_structure.iterChains():
        if chain.getChid() in gaps.keys():
            gaps_e = [x[0] for x in gaps[chain.getChid()]]
            gaps_b = [x[1] for x in gaps[chain.getChid()]]
        else:
            gaps_e = []
            gaps_b = []
        initial_residue, final_residue = FindInitialAndFinalResidues(chain)
        residues2ignore = []
        for residue in chain.iterResidues():
            resname = residue.getResname().strip()
            resnum = residue.getResnum()
            residue_atomnames = list(residue.getNames())
            if resname in supported_metals:
                continue
            if resname in supported_aminoacids:
                if charge_terminals:
                    if resnum == initial_residue or resnum in gaps_b:
                        zmatrix = ZMATRIX(resname + 'B')
                    elif resnum == final_residue or resnum in gaps_e:
                        zmatrix = ZMATRIX(resname + "E")
                    else:
                        zmatrix = ZMATRIX(resname)
                else:
                    if resname == "NMA":
                        zmatrix = ZMATRIX(resname + "E")
                    elif resname == "ACE":
                        zmatrix = ZMATRIX(resname + "B")
                    else:
                        zmatrix = ZMATRIX(resname)
            else:
                try:
                    zmatrix = ZMATRIX(resname)
                except:
                    print("UP {}".format(resname))
            if zmatrix.Name is None:
                print("  * The residue {} {} doesn't have a template, so it won't be checked.".format(resname, resnum))
                residues_without_template.append([resname, resnum, chain.getChid()])
                continue
            residue_atomnames.sort()
            if sorted(zmatrix.AtomNames) != residue_atomnames:
                atoms2delete = []
                atoms2add = []
                atoms2modify = []
                missing_atoms = [atom_name for atom_name in zmatrix.AtomNames if atom_name not in residue_atomnames]
                over_atoms = [atom_name for atom_name in residue_atomnames if atom_name not in zmatrix.AtomNames]
                if over_atoms:
                    if charge_terminals:
                        if resnum == final_residue or resnum in gaps_e:
                            if "HXT" in over_atoms:
                                over_atoms.pop(over_atoms.index('HXT'))
                                if 'OXT' in missing_atoms:
                                    atoms2modify.append(['HXT', 'OXT'])  # This changes the atom name
                                    missing_atoms.pop(missing_atoms.index('OXT'))
                                else:
                                    atoms2delete.append('HXT')
                        elif resnum == initial_residue or resnum in gaps_b:
                            if "H1" in over_atoms:
                                over_atoms.pop(over_atoms.index('H1'))
                            if "H2" in over_atoms:
                                over_atoms.pop(over_atoms.index('H2'))
                        if over_atoms:
                            print("   The residue {} {} has more atoms than the zmatrix." \
                                  " The extra atoms are:{}\n PELE won't work, review " \
                                  "them".format(resname, resnum, ",".join(over_atoms)))
                    else:
                        if resnum == initial_residue or resnum in gaps_b:
                            if "H1" in over_atoms:
                                over_atoms.pop(over_atoms.index('H1'))
                                if 'H' in missing_atoms:
                                    missing_atoms.pop(missing_atoms.index('H'))
                                    atoms2modify.append(['H1', 'H'])  # This changes the atom name
                                else:
                                    atoms2delete.append('H1')
                            if "H2" in over_atoms:
                                over_atoms.pop(over_atoms.index('H2'))
                                atoms2delete.append('H2')
                            if "H3" in over_atoms:
                                over_atoms.pop(over_atoms.index('H3'))
                                atoms2delete.append('H3')
                        elif resnum == final_residue or resnum in gaps_e:
                            if "HXT" in over_atoms:
                                over_atoms.pop(over_atoms.index('HXT'))
                                atoms2delete.append('HXT')
                            if "OXT"in over_atoms:
                                over_atoms.pop(over_atoms.index('OXT'))
                                atoms2delete.append('OXT')
                        if over_atoms:
                            print("   The residue {} {} has more atoms than the zmatrix. The extra atoms are:{}\n " \
                                  "PELE won't work, review them".format(resname, resnum, ",".join(over_atoms)))
                    # residues2fix["{} {} {}".format(resname, resnum, residue.getChid())] = {'delete': atoms2delete}
                    if debug:
                        print('over_atoms: ', over_atoms)
                        print("ZMATRIX: {}\n zmat atoms:{}".format(zmatrix.Name, sorted(zmatrix.AtomNames)))
                        print(" resi atoms:", sorted(residue_atomnames))
                if missing_atoms:
                    if "CA" in missing_atoms or "N" in missing_atoms or "O" in missing_atoms or "C" in missing_atoms:
                        print("  The residue {} {} is missing one or more heavy atoms in the backbone. " \
                              "It cannot be fixed.".format(resname, resnum, residue.getChid()))
                        if remove_missing_ter:
                            if resnum in [initial_residue, final_residue]:
                                print(" INFO: In structure {} the residue {} {} in chain {} will be eliminated".format(
                                    initial_structure.getTitle(), residue.getResname(),
                                    resnum, residue.getChid()))
                                try:
                                    residues2remove[residue.getChid()]
                                except KeyError:
                                    residues2remove[residue.getChid()] = [resnum]
                                else:
                                    residues2remove[residue.getChid()].append(resnum)
                                if resnum == initial_residue:
                                    next_res_resnum = resnum + 1
                                    next_residue = chain.getResidue(next_res_resnum)
                                    next_res_resname = next_residue.getResname()
                                    if charge_terminals:
                                        if next_res_resname != "PRO":
                                            atom = next_residue.getAtom("H")
                                            if atom is None:
                                                atoms2add = {"H1", "H2", "H3"}
                                            else:
                                                residues2ignore.append(next_res_resnum)
                                                atom.setName("H1")
                                                atoms2add = {"H2", "H3"}
                                        else:
                                            atom = residue.getAtom("H")
                                            if atom is None:
                                                atoms2add = {"H1", "H2"}
                                            else:
                                                atom.setName("H1")
                                                atoms2add = {"H2"}
                                        key = "{} {} {}".format(next_res_resname, next_res_resnum, residue.getChid())
                                        residues2fix[key]['add'] = atoms2add
                                        residues2fix[key]['delete'] = []
                                        residues2fix[key]['modify'] = []
                                elif resnum == final_residue:
                                    prev_res_resnum = resnum - 1
                                    prev_residue = chain.getResidue(prev_res_resnum)
                                    if prev_residue is None:
                                        print("There's no residue with number {} in chain {}".format(prev_res_resnum,
                                                                                                     chain.getChid()))
                                    elif charge_terminals:
                                        prev_res_resname = prev_residue.getResnames()[0]
                                        atoms2add = {"OXT"}
                                        key = "{} {} {}".format(prev_res_resname, prev_res_resnum, residue.getChid())
                                        try:
                                            residues2fix[key]
                                        except KeyError:
                                            residues2fix[key]['add'] = atoms2add
                                            residues2fix[key]['delete'] = []
                                            residues2fix[key]['modify'] = []
                                        else:
                                            residues2fix[key]['add'] += atoms2add
                        else:
                            print("  The program will be interrupted.".format(resname, resnum))
                            print("     This are the missing atoms: {}".format(missing_atoms))
                            sys.exit()
                    else:
                        if "H" in missing_atoms and resnum in gaps_b and resnum not in residues2ignore:
                            maestro_terminal_H = ["H1", "H2", "H11", "H22"]
                            possible_atoms2change = [atom_name for atom_name in maestro_terminal_H
                                               if atom_name in residue.getNames()]
                            if possible_atoms2change:
                                atomname_to_use = [0]
                                if atomname_to_use:
                                    atom = residue.getAtom(atomname_to_use)
                                    atom.setName('H')
                                    missing_atoms.pop(missing_atoms.index('H'))
                        if resname == "CYS":
                            key = "{}_{}".format(residue.getChid(), resnum)
                            if key in crosslinked_cysteines + charged_cysteines:
                                if key in charged_cysteines:
                                    residue.setResname('CYT')
                                elif "HG" in missing_atoms:
                                    missing_atoms.pop(missing_atoms.index('HG'))
                        for atom_name in missing_atoms:
                            if atom_name[0] != "H" and atom_name[0].isalpha() and atom_name != 'OXT':
                                print("  The residue {} {} is missing the heavy atom {}.\n" \
                                      "  All the atoms depending on this atom will" \
                                      " be placed according to the zmatrix.".format(resname, resnum, atom_name))
                                atoms2add = zmatrix.GetAllChildrenAtoms(atom_name)
                                atoms2add = set(atoms2add).union(set(missing_atoms))
                            elif resnum not in residues2ignore:
                                atoms2add = set(missing_atoms)
                key = "{} {} {}".format(resname, resnum, residue.getChid())
                try:
                    residues2fix[key]
                except KeyError:
                    if atoms2add or atoms2delete or atoms2modify:
                        residues2fix[key] = {'add': atoms2add, 'delete': atoms2delete, 'modify': atoms2modify}
                else:
                    residues2fix[key]['add'] += atoms2add,
                    residues2fix[key]['delete'] += atoms2delete
                    residues2fix[key]['modify'] += atoms2modify
    # print residues2fix, residues2remove
    return residues2fix, residues2remove, metals2coordinate, residues_without_template


def CheckMapAndZmatrix(zmap_atoms, mutation_map, mutation, residue):
    """This function checks that the number of atoms in the residue to mutate agree with the number of atoms in the map.
    It also checks the agreement between the mutation map and the zmatrix"""

    common_atoms = mutation_map[0]
    changing_atoms = mutation_map[1]
    volatile_atoms = mutation_map[2]
    volatile_atoms_behaviour, initial_atoms_index, final_atoms_index = mutation_map[3][
        "-".join([mutation['ini_resname'], mutation['fin_resname']])]

    # A small check to see that the map and the residue agree in the number of atoms.
    map_number_of_atoms = len(common_atoms) + len(changing_atoms) + len(volatile_atoms)
    residue2mutate_number_of_atoms = residue.numAtoms()
    map_atoms = common_atoms + [atom_pair[final_atoms_index] for atom_pair in changing_atoms]

    initial_residue_map_atoms = common_atoms + [atom_pair[initial_atoms_index] for atom_pair in changing_atoms]
    if volatile_atoms_behaviour == 'disappear':
        initial_residue_map_atoms += volatile_atoms
    else:
        map_atoms += volatile_atoms
    if map_number_of_atoms != residue2mutate_number_of_atoms and map_number_of_atoms != residue2mutate_number_of_atoms + len(
            volatile_atoms):
        print("Check the number of atoms of the structure because it doesn't match with the map.")
        print("Remember that mutations at the beginning or ending of the protein are not supported yet.")
        print("residue atoms: {}".format(set(residue.getNames())))
        print("initial residue map atoms: {}".format(initial_residue_map_atoms))
        sys.exit()

    if volatile_atoms_behaviour == 'appear':
        map_atoms += volatile_atoms
    if set(map_atoms) != set(zmap_atoms):
        print("The zmatrix for the desired aminoacid and the map for the mutation don't agree. Check Them.")
        print("map atoms: {}".format(set(map_atoms)))
        print("zmatrix atoms: {}".format(set(zmap_atoms)))
        sys.exit()


def CheckMutation(wt_structure, mutation):
    """
     This function checks whether the specified mutation is possible to make or not. And specifies the reason why it
    doesn't work.
     It may not be possible to make it due to the fact that the specified residue isn't present in the protein, or
    the final aminoacid to generate maybe isn't supported by the program.
    This function has as input the:
        - wt_structure: the structure to mutate
        - mutation: the mutation to do with the format ["original_resname", "resnum", "final_resname"]
        - valid_resnames_file: the path to the file containing a list with the supported mutations
    """

    if mutation['fin_resname'] not in supported_aminoacids:
        return False

    testing_the_resnum = "resnum {}".format(mutation['resnum'])
    testing_the_resname = "resname {}".format(mutation['ini_resname'])

    if wt_structure.select(testing_the_resnum) is None:
        print("The protein doesn't have a residue numbered {}".format(mutation['resnum']))
        return False
    elif wt_structure.select(testing_the_resname) is None:
        print("The protein doesn't have a residue named {}".format(mutation['ini_resname']))
        # print wt_structure.getResnames()
        return False
    elif mutation['chain']:
        testing_the_chain = "chain `{}` ".format(mutation['chain'])
        if wt_structure.select(testing_the_chain) is None:
            print("The protein doesn't have a chain named {}".format(mutation['chain']))
        elif wt_structure.select(testing_the_resname + ' and ' + testing_the_resnum + ' and ' + testing_the_chain) is None:
            print("The protein doesn't have the residue {0[ini_resname]} {0[chain]} {0[resnum]}".format(mutation))
            return False
    else:
        return True


def CheckClashes(mutated_protein, mutation, zmatrix, initial_residue,
                 final_residue, overlapfactor=0.7):
    """
    :param initial_residue:
    :param mutation:
    :rtype : dictionary
    :param mutated_protein: prody AtomGroup
    :param zmatrix: ZMATRIX
    :param overlapfactor: float

    A function to check the clashes between a given residue and the rest of the
    structure. It needs the number of the mutated residue the structure, the
    zmatrix of the mutated residue and the overlapfactor.

    The mutated protein is an Atomgroup object from prody that has the structure
    to check for clashes.

    The num_mutated_residue is the number of the mutated residue to check for
    clashes with the rest of the structure.

    The zmatrix contains the information extracted from the template of PELE.

    The overlapfactor by default is 0.7 but it can be set by the user to any
    number between 0 and 1. If it's one the atoms can be completely superimposed
    if the number is 0 they can't superimpose at all.
    """
    max_vdw = max(zmatrix.VdWRadius)
    contacts_object = Contacts(mutated_protein)
    if mutation["chain"]:
        mutated_residue = mutated_protein.select("resnum {0[resnum]} and chain {0[chain]} and not hydrogen".format(mutation))
    else:
        mutated_residue = mutated_protein.select("resnum {0[resnum]} and not hydrogen".format(mutation))
    mutated_residue_indices = mutated_residue.getIndices()
    clashes2check = {}
    for at in mutated_residue.iterAtoms():
        atom_name = at.getName()
        if atom_name in fixed_atoms:
            continue
        try:
            at_vdw = zmatrix.VdWRadius[zmatrix.AtomNames.index(atom_name)]
        except ValueError:
            print("The atom {} in the mutated residue doesn't agree with the Zmatrix.".format(atom_name))
            print(mutated_residue.getNames())
            print(zmatrix.AtomNames)
            raise ValueError
        dist = (at_vdw + max_vdw) * overlapfactor
        at_contacts = contacts_object.select(dist, at.getCoords())
        if at_contacts is None:
            continue
        possible_clashes = [clashing_atom.getIndex() for clashing_atom in at_contacts
                            if clashing_atom.getIndex() not in mutated_residue_indices]
        if possible_clashes:
            clashes2check[atom_name] = possible_clashes

    if mutated_residue.getResnames()[0] == "PRO":
        try:
            cd_atom_clashes = clashes2check["CD"]
        except KeyError:
            pass
        else:
            n_atom_index = mutated_residue.select("name N").getIndices()[0]
            try:
                cd_atom_clashes.index(n_atom_index)
            except ValueError:
                pass
            else:
                cd_atom_clashes.pop(cd_atom_clashes.index(n_atom_index))
    real_clashes = {}
    for key, indices in clashes2check.iteritems():
        real_clashing_indices = []
        for ind in indices:
            atom2check = mutated_protein.select('index {}'.format(ind))
            atom2check_resname = atom2check.getResnames()[0]
            atom2check_resnum = atom2check.getResnums()[0]
            if atom2check_resnum == initial_residue and atom2check_resname in supported_aminoacids:
                atom2check_zmatrix = ZMATRIX(atom2check_resname + 'B')
            elif atom2check_resnum == final_residue:
                atom2check_zmatrix = ZMATRIX(atom2check_resname + 'E')
            else:
                atom2check_zmatrix = ZMATRIX(atom2check_resname)
            if atom2check_zmatrix.Name is None:
                print("Cannot load the zmatrix for the residue {} {}.\n" \
                      "It won't be checked for clashes.".format(atom2check_resname, atom2check_resnum))
                continue
            atom2check_vdw_radius = atom2check_zmatrix.VdWRadius[
                atom2check_zmatrix.AtomNames.index(atom2check.getNames()[0])]
            key_vwd_radius = zmatrix.VdWRadius[zmatrix.AtomNames.index(key)]
            min_distance = (atom2check_vdw_radius + key_vwd_radius) * overlapfactor
            real_distance = calcDistance(mutated_residue.select('name {}'.format(key)), atom2check)
            if real_distance < min_distance:
                real_clashing_indices.append(ind)
        if real_clashing_indices:
            real_clashes[key] = real_clashing_indices

    return real_clashes


def CheckforGaps(structure, max_bond_distance):
    """
    This function checks all the residues in the protein and checks whether the N of the current residue and the
    C of the previous one are within peptidic bond distance (specified by the max_bond_distance variable)
    in order to check for gaps.
    :param structure: a prody object containing the structure to study.
    :param max_bond_distance: a float specifying the max bond distance to accept.
    :return: two dictionaries both containing the chain as key and the residues numbers (previous, current) involved
    in a bond as values, the first dictionary contains the information about the gaps and the second one contains
    the information about those not involved in gaps.
    """
    gaps = {}
    not_gaps = {}
    if max_bond_distance < 1.55:
        max_bond_distance = 1.55
    for chain in structure.iterChains():
        # initial, final = FindInitialAndFinalResidues(chain)
        residues = [residue for residue in chain]
        previous_residue_number = None
        for i, residue in enumerate(chain.iterResidues()):
            chain_id = chain.getChid()
            # print residue.getResnum() < previous_residue_number
            if previous_residue_number is not None:
                # if residue.getResnum() > previous_residue_number + 1:
                if residue.getResname() in supported_aminoacids:
                    if residue.getResname() == 'ACE':
                        try:
                            gaps[chain_id]
                        except KeyError:
                            gaps[chain_id] = []
                        gaps[chain_id].append([previous_residue_number, residue.getResnum()])
                    else:
                        current_residue_N = residue.getAtom("N")
                        previous_residue = chain.getResidue(previous_residue_number)
                        if not previous_residue:
                            previous_residue = residues[i-1]
                        previous_residue_C = previous_residue.getAtom('C')
                        if current_residue_N is not None and previous_residue_C is not None:
                            distance = calcDistance(current_residue_N, previous_residue_C)
                            if distance < max_bond_distance:
                                try:
                                    not_gaps[chain_id]
                                except KeyError:
                                    not_gaps[chain_id] = []
                                not_gaps[chain_id].append([previous_residue_number, residue.getResnum()])
                            else:
                                try:
                                    gaps[chain_id]
                                except KeyError:
                                    gaps[chain_id] = []
                                gaps[chain_id].append([previous_residue_number, residue.getResnum()])
                        elif current_residue_N is None:
                            print("   * There's a problem with residue {} {} {} it" \
                                  " doesn't have the N atom".format(residue.getResname(),
                                                                    residue.getResnum(),
                                                                    residue.getChid()))
                        elif previous_residue_C is None:
                            print("   * There's a problem with residue {} {} {} it doesn't have the C atom".format(
                                previous_residue.getResname(),
                                previous_residue.getResnum(),
                                previous_residue.getChid()))
                if residue.getResnum() == previous_residue_number:
                    if residue.getIcode() == '':
                        if residue.hetero is not None:
                            residue.setResnum(previous_residue_number + 1)
                        else:
                            print('   * WARNING: There are two residues with the same resnum in the same chain.' \
                                  'The residue {} {} {} '.format(residue.getChid(),
                                                                 residue.getResnum(),
                                                                 residue.getResname()))
                elif residue.getResnum() < previous_residue_number + 1:
                    if residue.hetero is not None:
                        residue.setResnum(previous_residue_number + 1)
                    else:
                        # print residue.getResnum(), previous_residue_number
                        print("   * WARNING! The next residue has a lower resnum than the previous one in the chain??")
                        print("    * CHAIN: {} RESIDUE: {}  {} {}".format(residue.getChid(),
                                                                    residue.getResnum(),
                                                                    residue.getChid(),
                                                                    residue.getResname()))
                        print("    * previous: RESIDUE {} {}  {}".format(previous_residue.getResname(),
                                                                   previous_residue.getResnum(),
                                                                   previous_residue.getChid()))
                        sys.exit('This should never happen, please review your structure.')
                else:
                    pass
            previous_residue_number = residue.getResnum()
    return gaps, not_gaps



import sys
from argparse import ArgumentParser, RawTextHelpFormatter
from re import match
from PPP.global_variables import supported_aminoacids, input_keywords, aminoacids_1letter, aminoacids_3letter
from PPP import parameters_help

from prody import parsePDB

__author__ = 'jelisa'


def ParseMutations(unprocessed_mutation, structure_filename):
    """"""
    # TODO: This fuction should be able to allow the user to specify
    # regions and selections in the input instead of just one residue.
    processed_mutations = []
    chain = ""
    for mutation in unprocessed_mutation:
        splitted_mutation = mutation.split()
        if "residue" not in mutation:
            if len(splitted_mutation) == 4:
                original_resname, resnum, chain, final_resname = [value.upper() for value in splitted_mutation]
            elif len(splitted_mutation) == 3:
                original_resname, resnum, final_resname = [value.upper() for value in splitted_mutation]
            else:
                selection_string_raw, final_resname_raw = mutation.split("TO")
                structure = parsePDB(structure_filename)
                #TODO: Talk with fatima about the selections they'll be doing.
                if "within" in selection_string_raw:
                    selection_center = selection_string_raw.split("of")[-1]
                    selection_string = "({}) and (not {}) and (not resname HOH)".format(selection_string_raw,
                                                                                        selection_center)
                    mutations2add = []
                    selection = structure.select(selection_string).copy()
                    if selection is None:
                        print("Something is wrong with the selection string it isn't selecting anything. Review it!")
                        print(" Selection string: {}".format(selection_string))
                        print("The program will be interrupted.")
                        sys.exit()
                    for residue in selection.iterResidues():
                        original_resname = residue.getResname()
                        resnum = residue.getResnum()
                        chain = residue.getChid()
                        final_resname = final_resname_raw
                        # mutations2add.append(mutation)
                else:
                    selection = selection_string_raw.split()
                    original_resname = selection[selection.index("resname") + 1].upper()
                    resnum = selection[selection.index("resnum") + 1]
                    if "chain" in selection:
                        chain = selection[selection.index("chain") + 1].upper()
                    else:
                        chain = ""
                    final_resname = final_resname_raw.strip().upper()
        else:
            if len(splitted_mutation) != 4:
                try:
                    int(splitted_mutation[splitted_mutation.index("residue") + 2])
                except ValueError:
                    try:
                        int(splitted_mutation[splitted_mutation.index("residue") + 1])
                    except ValueError:
                        print("Your mutation string is wrong." \
                              " It should look like: 'residue XXX N to YYY' with N being the resnum and XXX, YYY being supported aminoacids")
                        sys.exit()
                    else:
                        original_resname = splitted_mutation[splitted_mutation.index("residue") + 2].upper()
                        resnum = splitted_mutation[splitted_mutation.index("residue") + 1]
                else:
                    original_resname = splitted_mutation[splitted_mutation.index("residue") + 1].upper()
                    resnum = splitted_mutation[splitted_mutation.index("residue") + 2]
                try:
                    chain = splitted_mutation[splitted_mutation.index("chain") + 1]
                except ValueError:
                    if splitted_mutation[splitted_mutation.index("residue") + 3] != "to":
                        chain = splitted_mutation[splitted_mutation.index("residue") + 3]
                    else:
                        chain = ""
                final_resname = splitted_mutation[splitted_mutation.index('to') + 1].upper()
        mutation = {'ini_resname': original_resname, "resnum": resnum,
                    "chain": chain, "fin_resname": final_resname}
        if mutation["fin_resname"] in supported_aminoacids:
            processed_mutations.append(mutation)
            print('a')
        elif mutation["fin_resname"].lower() in input_keywords.keys():
            keyword = mutation["fin_resname"].lower()
            histidines_equivalents = ["HID", "HIE", "HIP", "HIS"]
            if original_resname in histidines_equivalents:
                exclude_histidines = True
            else:
                exclude_histidines = False
            for aminoacid in input_keywords[keyword]:
                if exclude_histidines and aminoacid in histidines_equivalents:
                    continue
                elif aminoacid == original_resname:
                    continue
                else:
                    mutation = {'ini_resname': original_resname, "resnum": resnum,
                                "chain": chain, "fin_resname": aminoacid}
                    processed_mutations.append(mutation)
        else:
            print('c', mutation["fin_resname"], input_keywords.keys())
    return processed_mutations


def ParseArguments():
    parser = ArgumentParser(description=parameters_help.program_description,
                            formatter_class=RawTextHelpFormatter)
    parser.add_argument('-input_pdb', '-ipdb', required=True, nargs='*', help=parameters_help.ipdb_description)
    parser.add_argument('-output_pdb', '-opdb', default=[], nargs="+", help=parameters_help.opdb_description)
    parser.add_argument('-mutation', '-mut', type=str, default='', nargs='+',
                        help=parameters_help.mutation_description)
    parser.add_argument("-mutants_from_file", '-mut_file', default=False,
                        help=parameters_help.mutants_from_file_description)
    parser.add_argument("-mutant_multiple", action="store_true",
                        help=parameters_help.mutant_multiple_description)
    parser.add_argument("-charge_terminals", action="store_true",
                        help=parameters_help.charge_terminals_description)
    parser.add_argument("-no_gaps_ter", action="store_true", default=False,
                        help=parameters_help.no_gaps_ter_description)
    parser.add_argument("-pdb_resolution", default=2.5, type=float,
                        help=parameters_help.pdb_resolution_description)
    parser.add_argument("-make_unique", default=False,
                        help=parameters_help.make_unique_description)
    parser.add_argument("-remove_terminal_missing", default=False, action="store_true",
                        help=parameters_help.remove_terminal_missing_description)
    # parser.add_argument('-force_field', '-ff', type=str, default="OPLS2005",
    #                     choices=["OPLS2005", "OPLS2001", "AMBER99sbBSC0"],
    #                     help=parameters_help.ff_description)
    args = parser.parse_args()

    if len(args.input_pdb) > 1 and (args.mutation != '' or args.mutants_from_file):
        print("This options are incompatible, when you want to do mutations you should use only one structure at time.")
        return None
    if len(args.input_pdb) == 1:
        args.input_pdb = args.input_pdb[0]
    if args.mutation:
        args.mutation = ParseMutations(args.mutation, args.input_pdb)
    if args.mutants_from_file:
        filein = open(args.mutants_from_file, 'r')
        mutations_text = filein.readlines()
        filein.close()
        args.mutation = []
        args.output_pdb = []
        for line in mutations_text:
            line = line.strip()
            if line == '':
                continue
            output_file, mutation = line.split('\t')
            splitted_mutation = mutation.split()
            if len(splitted_mutation) == 4:
                args.mutation.append(
                    {'ini_resname': splitted_mutation[0].upper(), "resnum": splitted_mutation[1].upper(),
                     "chain": splitted_mutation[2].upper(), "fin_resname": splitted_mutation[3].upper()})
            elif len(splitted_mutation) == 3:
                args.mutation.append(
                    {'ini_resname': splitted_mutation[0].upper(), "resnum": splitted_mutation[1].upper(),
                     "fin_resname": splitted_mutation[3].upper()})
    output_files_number = len(args.output_pdb)
    if output_files_number == 0:
        args.output_pdb = []
        args.output_pdb.append(args.input_pdb[:-4] + "_processed" + args.input_pdb[-4:])
    elif output_files_number == 1:
        if len(args.mutation) > 1:
            general_name = args.mutation[0].split('.pdb')[0]
            for mutation in args.mutation:
                mutation_id = aminoacids_1letter[aminoacids_3letter.index(mutation["ini_resname"])]
                if mutation["chain"]:
                    mutation_id += mutation["chain"]
                mutation_id += mutation["resnum"]
                mutation_id += aminoacids_1letter[aminoacids_3letter.index(mutation["fin_resname"])]
                output_name = "{}_{}.pdb".format(general_name, mutation_id)
                args.output_pdb.append(output_name)


    return args


def FindInitialAndFinalResidues(structure):
    """
    This function finds the first and the last residue of a protein.
    """

    initial_residue, final_residue = 0, 0
    first_aa = False
    for res in structure.iterResidues():
        resnum = res.getResnum()
        resname = res.getResname()
        if resname in supported_aminoacids:
            if not first_aa:
                initial_residue = resnum
                first_aa = True
            elif resnum > final_residue:
                final_residue = resnum

    return initial_residue, final_residue


def PDBwriter(output_file_name, structure, make_unique, residues2remove, no_ter_mkar_for_gaps, no_proteic_ligand=None,
              gaps={}, no_gaps={}, mid_chain_nonstd_residue=[]):
    supported_aminoacids.extend(mid_chain_nonstd_residue)
    if '.pdb' not in output_file_name:
        output_file_name += '.pdb'
    outfile = open(output_file_name, 'w')
    alt_loc = " "  # It should always be an empty string.
    insertion_code = " "  # It should always be an empty string.
    occupancy = 1.0  # It should be always 1
    serial = 1
    for chain in structure.iterChains():
        ter = False
        resnums = [residue.getResnum() for residue in chain]
        if chain.getChid() in gaps.keys():
            gaps_e = [x[0] for x in gaps[chain.getChid()]]
        else:
            gaps_e = []
        if chain.getChid() == make_unique and no_proteic_ligand is not None:
            change_names = True
        else:
            change_names = False
        ligand_possible_atoms = {"C": 1, "N": 1, "O": 1, "H": 1, "F": 1,
                                 "S": 1, "P": 1, "BR": 1, "I": 1, "CL": 1,
                                 "FE": 1, "CA": 1}

        for index, residue in enumerate(chain.iterResidues()):
            ter = False
            resnum = residue.getResnum()
            eliminate_hxt = False
            eliminate_h = False
            if residue.getChid() in residues2remove.keys():
                if resnum in residues2remove[residue.getChid()]:
                    continue
            if resnum in gaps_e:
                ter = True
                if no_ter_mkar_for_gaps:
                    ter = False
                if " HXT" in residue.getNames():
                    eliminate_hxt = True
                elif " H1 " in residue.getNames() or " H2 " in residue.getNames():
                    eliminate_h = True
            chain_id = residue.getChid()
            resname = residue.getResname()
            if resname == "NMA":
                ter = True
                atom_hetatm = "ATOM"
            elif resname in supported_aminoacids:
                atom_hetatm = "ATOM"
                try:
                    next_residue = chain.select("resnum `{}`".format(resnums[index+1]))
                except IndexError:
                    pass
                else:
                    if next_residue.getResnames()[0] not in supported_aminoacids:
                        ter = True
            else:
                atom_hetatm = "HETATM"
                ter = True
            for atom in residue.iterAtoms():
                atom_name = atom.getName()
                if atom_name == ' HXT' and eliminate_hxt:
                    continue
                elif atom_name in [' H1 ', ' H2 '] and eliminate_h:
                    continue
                if change_names:
                    if "Ca" in atom_name:
                        raw_atom_name = atom_name.strip()[:2]
                    elif "CA" in atom_name:
                        if atom_name[:2] == "CA":
                            raw_atom_name = atom_name.strip()[:2]
                        else:
                            raw_atom_name = atom_name.strip()[0]
                    else:
                        raw_atom_name = atom_name.strip()
                        if len(raw_atom_name) > 1:
                            # print "'{}'".format(atom_name)
                            raw_atom_name = raw_atom_name.upper()
                            if raw_atom_name[:2] in ligand_possible_atoms.keys():
                                raw_atom_name = raw_atom_name[:2]
                            elif raw_atom_name not in ligand_possible_atoms.keys():
                                raw_atom_name = raw_atom_name[0]
                                if raw_atom_name not in ligand_possible_atoms.keys():
                                    print("INVALID ATOM NAME in the ligand file: {}".format(raw_atom_name))
                        elif raw_atom_name not in ligand_possible_atoms.keys():
                            print("INVALID ATOM NAME in the ligand file: {}".format(raw_atom_name))
                    atom_name = raw_atom_name + str(ligand_possible_atoms[raw_atom_name])
                    if len(atom_name) == 1:
                        formated_atom_name = " " + atom_name + "  "
                    elif len(atom_name) == 2:
                        formated_atom_name = " " + atom_name + " "
                    elif len(atom_name) == 3:
                        formated_atom_name = " " + atom_name
                    ligand_possible_atoms[raw_atom_name] += 1
                else:
                    formated_atom_name = atom_name
                x, y, z = atom.getCoords()
                element = atom.getElement()
                segment = atom.getSegname()
                b_factor = atom.getBeta()
                if atom.getCharge() is None:
                    charge = " "
                else:
                    charge = atom.getCharge()
                pdbstring = "{1:6}{2:5}{0:1}{3:4}{4:1}{5:3}{0:1}{6:1}{7:4}{8:1}{0:3}" \
                            "{9:8.3f}{10:8.3f}{11:8.3f}{12:6.2f}{13:6.2f}{0:6}{14:<4}{15:>2}{16:2}\n"
                pdbstring = pdbstring.format('', atom_hetatm, serial, formated_atom_name, alt_loc,
                                             resname, chain_id, resnum, insertion_code, x, y,
                                             z, occupancy, b_factor, segment,
                                             element, charge)
                outfile.write(pdbstring)
                serial += 1
            if ter:
                outfile.write("TER\n")
        if not ter:
            outfile.write("TER\n")
    outfile.close()
    return


def RenumberStructure(initial_structure, debug=False):
    renumbered_structure = initial_structure.copy()
    for chain in renumbered_structure.iterChains():
        residue_number = 1
        chain_id = chain.getChid()
        for residue in chain.iterResidues():
            if residue.getResnum() == debug:
                print(residue.getResnum(), residue.getResname(), residue_number)
            residue.setResnum(residue_number)
            residue.setIcode('')
            residue_number += 1
    return renumbered_structure



import sys

import numpy as np
from prody import writePDB, parsePDB, AtomGroup
from PPP.checks_module import CheckMapAndZmatrix
from PPP.coordinates_module import GenerateCoordinatesFromZmatrix, ModifyCoords, ModifyCoordinatesPRO
from PPP.global_variables import default_mutations_maps
from PPP.program_own_classes import ZMATRIX

__author__ = 'jelisa'

# Defining the defaults values for the files thet define the valid maps
# This file should have the format of a python list: ["XXX1", "XXX2", ... , "XXXn"]


def Readmaps(mutation):
    mutation_arbitrary_order = [mutation['ini_resname'], mutation["fin_resname"]]
    mutation_arbitrary_order.sort()
    mutation_maps_key = "-".join(mutation_arbitrary_order)
    try:
        mutation_map = default_mutations_maps[mutation_maps_key]
    except KeyError:
        print("Something is wrong, the mutation you're asking for is not present in the mutations maps.")
        sys.exit()
    return mutation_maps_key, mutation_map


def EliminateAtoms(initial_residue, atoms2eliminate):
    """This function gets a residue and a list of atomnames to eliminate from the residue.
    It returns the residue without the atoms."""
    atoms2eliminate = " OR ".join(atoms2eliminate)
    mutated_residue = initial_residue.select('not (name {})'.format(atoms2eliminate)).copy()
    return mutated_residue


def AddAtoms(initial_residue, atoms2add, atomnames_of_2_letters, residue_data, zmatrix, verbose=True):
    """
    This function returns a mutated residue with the atoms added.
    This function gets a residue and a list of atomnames to add to the mutation.
    :rtype : object
    :param initial_residue:
    :param atoms2add: 
    :param atomnames_of_2_letters: 
    :param residue_data: 
    :param zmatrix: 
    :param verbose: 
    """
    if verbose:
        print('Adding new atoms:')
    new_atoms_elements = []
    for atom in atoms2add:
        if verbose:
            print(" # {}".format(atom))
        if atom[:2] in atomnames_of_2_letters:
            new_atoms_elements.append(atom[:2])
        else:
            new_atoms_elements.append(atom[0])

    new_atoms = AtomGroup('New atoms')
    number_of_new_atoms = len(atoms2add)
    new_atoms.setNames(list(atoms2add))
    new_atoms.setElements(new_atoms_elements)
    new_atoms.setResnames([residue_data['fin_resname']] * number_of_new_atoms)
    new_atoms.setResnums([residue_data['resnum']] * number_of_new_atoms)
    new_atoms.setChids([residue_data['chain']] * number_of_new_atoms)
    new_atoms.setAltlocs([''] * number_of_new_atoms)
    new_atoms.setBetas([0] * number_of_new_atoms)
    new_atoms.setIcodes([''] * number_of_new_atoms)
    new_atoms.setOccupancies([1] * number_of_new_atoms)
    new_atoms.setSegnames([''] * number_of_new_atoms)
    new_atoms.setSerials([0] * number_of_new_atoms)
    temporary_coords = np.ones([len(atoms2add), 3])
    new_atoms.setCoords(temporary_coords)
    incomplete_residue = initial_residue + new_atoms
    for atom in new_atoms.iterAtoms():
        atom_new_coordinates = GenerateCoordinatesFromZmatrix(incomplete_residue, [atom.getName()],
                                                                                 zmatrix)
        atom.setCoords(atom_new_coordinates)
        incomplete_residue = initial_residue + new_atoms
    if verbose:
        print('Done')
    return initial_residue + new_atoms


def ModifyExistingAtoms(initial_residue, atoms2change, atomnames_of_2_letters, initial_atoms_index, final_atoms_index,
                        zmatrix):
    residue2modify = initial_residue.copy()
    print("    * Modifying the atoms:")
    for at_pair in atoms2change:
        ini_atomname = at_pair[initial_atoms_index]
        fin_atomname = at_pair[final_atoms_index]
        print("      # {} -> {}".format(at_pair[initial_atoms_index], at_pair[final_atoms_index]))
        element = fin_atomname[:2]
        if element not in atomnames_of_2_letters:
            element = at_pair[final_atoms_index][0]
        atom2modify = residue2modify.select("name {0}".format(ini_atomname))
        atom2modify.setNames([fin_atomname])
        atom2modify.setElements([element])
        if fin_atomname == 'H':
            atom_coords = ModifyCoordinatesPRO(residue2modify, zmatrix, fin_atomname, fin_atomname)
        else:
            atom_coords = GenerateCoordinatesFromZmatrix(residue2modify, [fin_atomname], zmatrix)
        atom2modify.setCoords(atom_coords)
    print("      Done")
    return residue2modify


def ModifyFromToGLY(initial_residue, zmatrix, resname, atoms2change, initial_atoms_index, final_atoms_index, atomnames_of_2_letters):
    modified_residue = initial_residue.copy()
    initial_atoms, final_atoms = [], []
    print("Modifying the atoms:")
    for at_pair in atoms2change:
        initial_atoms.append(at_pair[initial_atoms_index])
        final_atoms.append(at_pair[final_atoms_index])
    for at_pair in atoms2change:
        initial_name = at_pair[initial_atoms_index]
        final_name = at_pair[final_atoms_index]
        print(" # {} -> {}".format(initial_name, final_name))
        atom2modify = modified_residue.select("name {}".format(initial_name))
        atom2modify.setNames([final_name])
        if final_name == "H":
            coords = ModifyCoordinatesPRO(modified_residue, zmatrix, final_name, final_name)
        else:
            coords = ModifyCoords(modified_residue, zmatrix, final_name)
        atom2modify.setCoords(coords)
        if final_name in atomnames_of_2_letters:
            pass
        else:
            atom2modify.setElements([final_name[0]])
        atom2modify.setBetas([0])
        atom2modify.setAltlocs([""])
    modified_residue.setResnames([resname] * modified_residue.numAtoms())
    return modified_residue


def AddHfromPRO(initial_residue, zmatrix, mutation):
    new_atom = AtomGroup('Hydrogen')
    new_atom.setNames(['H'])
    new_atom.setElements(['H'])
    new_atom.setResnames([zmatrix.Name])
    new_atom.setResnums([mutation['resnum']])
    new_atom.setChids([mutation['chain']])
    new_atom.setAltlocs([''])
    new_atom.setBetas([0])
    new_atom.setIcodes([''])
    new_atom.setOccupancies([1])
    new_atom.setSegnames([''])
    new_atom.setSerials([0])
    new_atom.setCoords(ModifyCoordinatesPRO(initial_residue, zmatrix, 'H', 'CD'))
    return initial_residue + new_atom

def Mutate(wt_structure, mutation):
    """
    This is the main method that mutates the protein it calls all the other functions needed to do the mutation.
    It reads the mutation map, applies it to the residue, then checks the geometry and coordinates, and finally returns
    the mutated protein.
    """

    # Separate the residue from the rest of the protein, and the rest of the protein is divided in two the part before
    # the residue to mutate, and the part after it. To ease the union of the parts after the mutation process.
    if mutation['chain']:
        initial_residue = wt_structure.select("resname {0[ini_resname]} and chain {0[chain]} and resnum {0[resnum]}".format(mutation)).copy()
        if wt_structure.select("not chain {0[chain]}".format(mutation)) is not None:
            rest_of_the_structure = wt_structure.select("not chain {0[chain]}".format(mutation)).copy()
        else:
            rest_of_the_structure = None
        initial_part_of_the_protein = wt_structure.select("resnum < {0[resnum]} and chain {0[chain]}".format(mutation)).copy()
        final_part_of_the_protein = wt_structure.select("resnum > {0[resnum]} and chain {0[chain]}".format(mutation)).copy()
    else:
        initial_residue = wt_structure.select("resname {0[ini_resname]} and resnum {0[resnum]}".format(mutation)).copy()
        rest_of_the_structure = ''
        initial_part_of_the_protein = wt_structure.select("resnum < {0[resnum]}".format(mutation)).copy()
        final_part_of_the_protein = wt_structure.select("resnum > {0[resnum]}".format(mutation)).copy()


    # This block reads the map and extracts the necessary information
    mutation_maps_key, mutation_map = Readmaps(mutation)
    changing_atoms = mutation_map[1]
    volatile_atoms = mutation_map[2]
    volatile_atoms_behaviour, initial_atoms_index, final_atoms_index = mutation_map[3][
        "-".join([mutation['ini_resname'], mutation['fin_resname']])]

    zmatrix = ZMATRIX(mutation['fin_resname'])
    if zmatrix.Name is None:
        print("The template for the final residue doesn't exist. Check your environment parameters.")
        sys.exit()
    CheckMapAndZmatrix(zmatrix.AtomNames, mutation_map, mutation, initial_residue)
    atoms2check_for_clashes = [atom_pair[final_atoms_index] for atom_pair in changing_atoms]

    # This handles the change in the element and atom name.
    atomnames_of_2_letters = ["FE"]  # Modify this list to add more elements that should use 2 letters

    if "GLY" in mutation.values():
        if volatile_atoms_behaviour == "disappear":
            residue2mutate = EliminateAtoms(initial_residue, volatile_atoms)
            mutated_residue = ModifyFromToGLY(residue2mutate, zmatrix, mutation['ini_resname'], changing_atoms,
                                              initial_atoms_index, final_atoms_index, atomnames_of_2_letters)
        elif volatile_atoms_behaviour == "appear":
            residue2mutate = ModifyFromToGLY(initial_residue, zmatrix, mutation['fin_resname'], changing_atoms,
                                             initial_atoms_index, final_atoms_index, atomnames_of_2_letters)
            mutated_residue = AddAtoms(residue2mutate, volatile_atoms, atomnames_of_2_letters, mutation, zmatrix)

        elif volatile_atoms_behaviour == '' and volatile_atoms == []:
            pass
        else:
            print("Something is wrong with the appearing/disappearing value in the file with the maps. Check it!")
            print("The problem is in the entry for the mutation {} it has the value {}. It should be appear or disappear.".format(
                mutation_maps_key, volatile_atoms_behaviour))
            sys.exit()
        if mutation["fin_resname"] == "PRO":
            print('here')
            atoms2replace = ["CD", "HD2", "HD3"]
            for atom_name in atoms2replace:
                atom = mutated_residue.select('name {}'.format(atom_name))
                coords = GenerateCoordinatesFromZmatrix(mutated_residue, [atom_name], zmatrix)
                atom.setCoords(coords)
        mutated_residue.setResnames([zmatrix.Name] * mutated_residue.numAtoms())
    elif mutation['ini_resname'] == "PRO" and 'H' in volatile_atoms:
        residue2mutate = AddHfromPRO(initial_residue, zmatrix, mutation)
        residue2mutate = ModifyExistingAtoms(residue2mutate, changing_atoms, atomnames_of_2_letters,
                                             initial_atoms_index, final_atoms_index, zmatrix)
        mutated_residue = AddAtoms(residue2mutate, volatile_atoms[1:], atomnames_of_2_letters, mutation, zmatrix)
        mutated_residue.setResnames([zmatrix.Name] * mutated_residue.numAtoms())
    else:
        residue2mutate = ModifyExistingAtoms(initial_residue, changing_atoms, atomnames_of_2_letters,
                                             initial_atoms_index, final_atoms_index, zmatrix)
        residue2mutate.setResnames([mutation['fin_resname']] * initial_residue.numAtoms())
        # Eliminates or Creates the necessary atoms
        if volatile_atoms_behaviour == "disappear":
            mutated_residue = EliminateAtoms(residue2mutate, volatile_atoms)
        elif volatile_atoms_behaviour == "appear":
            mutated_residue = AddAtoms(residue2mutate, volatile_atoms, atomnames_of_2_letters, mutation, zmatrix)
            atoms2check_for_clashes += volatile_atoms
        elif volatile_atoms_behaviour == '' and volatile_atoms == []:
            mutated_residue = residue2mutate
        else:
            print("Something is wrong with the appearing/disappearing value in the file with the maps. Check it!")
            print("The problem is in the entry for the mutation {} it has the value {}. It should be appear or disappear.".format(
                mutation_maps_key, volatile_atoms_behaviour))
            sys.exit()
    if rest_of_the_structure:
        mutated_structure = initial_part_of_the_protein + mutated_residue + final_part_of_the_protein + rest_of_the_structure
    else:
        mutated_structure = initial_part_of_the_protein + mutated_residue + final_part_of_the_protein

    return mutated_structure, zmatrix


def main():
    """
    use like this:
    $> python mutational_module.py pdb_file.pdb 'residue XXX N to YYY'
    """
    initial_structure_path = sys.argv[1]
    mut = sys.argv[2].split()
    mutation = {'ini_resname': mut[1], 'resnum': mut[2], 'fin_resname': mut[4], }
    ff = "force_field parameters"

    initial_structure = parsePDB(initial_structure_path)

    mutated_protein, clashes = Mutate(initial_structure, mutation)
    writePDB('processed.pdb', mutated_protein)
    return mutated_protein, clashes


if __name__ == "__main__":
    mutated_residue, clashes = main()



import os

__author__ = 'jelisa'
DIR = os.path.dirname(__file__)

class ZMATRIX:
    """
    A class to store de data from the template file for PELE
    """

    AtomNames = None

    def __init__(self, aminoacid_name):
        # First the attributes are created as lists to be able to modify them when reading the data
        self.AtomNums = []
        self.Name = aminoacid_name
        self.AtomParents = []
        self.AtomNames = []
        self.BondLengths = []
        self.BondAngles = []
        self.Diherdrals = []
        self.VdWRadius = []
        self.ReadZmatrix(aminoacid_name)
        # The attributes are transformed to tuple so they can't be changed by mistake later on. It's really easy to
        # modify a list without noticing.
        self.AtomNums = tuple(self.AtomNums)
        self.AtomParents = tuple(self.AtomParents)
        self.AtomNames = tuple(self.AtomNames)
        self.BondLengths = tuple(self.BondLengths)
        self.BondAngles = tuple(self.BondAngles)
        self.Diherdrals = tuple(self.Diherdrals)
        self.VdWRadius = tuple(self.VdWRadius)

    def ReadZmatrix(self, aminoacid_name):
        """This function loads the zmatrix information for the desired aminoacid
        :rtype : a ZMATRIX object
        :param aminoacid_name: a string containing the 3 letters code for the desired aminoacid.
        """
        filename = "DeataLocal/Templates/OPLS2005/Protein/" + aminoacid_name.lower()
        try:
            filein = open(filename, 'r')
        except IOError:
            filename = "DaetaLocal/Templates/OPLS2005/HeteroAtoms/" + aminoacid_name.lower().strip() + 'z'
            try:
                filein = open(filename, 'r')
            except IOError:
                filename = os.path.join(DIR, "Data/Templates/OPLS2005/Protein/") + aminoacid_name.lower()
                try:
                    filein = open(filename, 'r')
                except IOError:
                    filename = os.path.join(DIR, "Data/Templates/OPLS2005/HeteroAtoms/") + aminoacid_name.lower() + 'z'
                    try:
                        filein = open(filename, 'r')
                    except IOError:
                        print("    * No such file or directory: {}".format(filename))
                        self.Name = None
                        return None
                    else:
                        self.Name = aminoacid_name.upper() + 'Z'
                else:
                    self.Name = aminoacid_name.upper()
            else:
                self.Name = aminoacid_name.upper() + 'Z'
        else:
            self.Name = aminoacid_name.upper()
        raw_data = filein.readlines()
        filein.close()
        add_lines = False
        add_vdw = False
        num_of_atoms = 0
        counter = 0
        for line in raw_data:
            if line.strip() == "":
                print("The template for the residue {} in {} has a blank line. " \
                      "Check it or not! Whatever you want.".format(self.Name, filename))
                continue
            if self.Name[-1] in ['z', 'b', 'e', 'a']:
                possible_name = line[:5]
            else:
                tmp_list = line.split()
                resname = tmp_list[0]
            if tmp_list[0] == self.Name:
                set_name = True
            else:
                possible_name = "".join(line[:5].strip().split())
                if possible_name == self.Name or possible_name == self.Name[:-1]:
                    set_name = True
                else:
                    set_name = False
            if set_name:
                num_of_atoms = int(tmp_list[1])
                counter = num_of_atoms
                add_lines = True
            elif add_lines:
                self.AtomNums.append(int(tmp_list[0]))
                self.AtomParents.append(int(tmp_list[1]))
                self.AtomNames.append(tmp_list[4].strip('_'))
                self.BondLengths.append(float(tmp_list[6]))
                self.BondAngles.append(float(tmp_list[7]))
                self.Diherdrals.append(float(tmp_list[8]))
                counter -= 1
                if counter == 0:
                    add_lines = False
            elif tmp_list[0] == "NBON":
                add_vdw = True
                counter = num_of_atoms
                if counter == 0:
                    print(" - There's a problem with the template of the residue {}.".format(aminoacid_name))
                    break
            elif add_vdw:
                self.VdWRadius.append((float(tmp_list[1]) / 2))
                counter -= 1
                if counter == 0:
                    break


    def GetParentName(self, atom2place):
        atom2place_position = self.AtomNames.index(atom2place)
        parent_num = self.AtomParents[atom2place_position]
        parent_position = self.AtomNums.index(parent_num)
        parent_name = self.AtomNames[parent_position]
        return parent_name

    def GetDirectChildrenAtoms(self, parent_atom_name):
        parent_atom_number = self.AtomNums[self.AtomNames.index(parent_atom_name)]
        if parent_atom_number in self.AtomParents:
            number_of_direct_children = self.AtomParents.count(parent_atom_number)
            children_indexes = [self.AtomParents.index(parent_atom_number)]
            for x in range(number_of_direct_children - 1):
                new_index = self.AtomParents[children_indexes[-1] + 1:].index(parent_atom_number)
                next_children = new_index + children_indexes[-1] + 1
                # The +1 is due to the fact that lists indexes start from 0
                children_indexes.append(next_children)
            children = [self.AtomNames[index] for index in children_indexes]
        else:
            children = []
        return children

    def GetAllChildrenAtoms(self, parent_atom_name, recursion_count=0):
        children = self.GetDirectChildrenAtoms(parent_atom_name)
        if recursion_count == 10000:
            raise RuntimeError("The function GetChildrenAtoms has arrived to the 10000 iterations!\n"
                               "             It's impossible that an aminoacid has so many atoms. Review it")
        for child in children:
            child_number = self.AtomNums[self.AtomNames.index(child)]
            if child_number in self.AtomParents:
                recursion_count += 1
                child_children = self.GetDirectChildrenAtoms(child)
                children.extend(child_children)
        return children

    def GetBrotherAtoms(self, atom_name):
        parent_name = self.GetParentName(atom_name)
        brothers = [brother for brother in self.GetDirectChildrenAtoms(parent_name)
                    if brother != atom_name and brother not in ['CA', 'C', 'N', 'O']]  # and brother[0] != 'H']
        return brothers

    def GetData2ComputeCoords(self, atom_name):
        at3_name = self.GetParentName(atom_name)
        at2_name = self.GetParentName(at3_name)
        at1_name = self.GetParentName(at2_name)
        atom_position_in_zmatrix = self.AtomNames.index(atom_name)
        r = self.BondLengths[atom_position_in_zmatrix]
        deta = self.BondAngles[atom_position_in_zmatrix]
        fi = self.Diherdrals[atom_position_in_zmatrix]
        return at1_name, at2_name, at3_name, r, deta, fi

    def ComputeDeltaFi(self):
        self.DeltaFi = [None for at in self.AtomNums]
        self.DeltaFiReference = [None for at in self.AtomNums]
        for index, atom in enumerate(self.AtomNames):
            if self.DeltaFi[index] is not None:
                continue
            elif atom in ["N", "H", "CA", "HA", "C", "O", "CB"]:
                self.DeltaFi[index] = 0
                continue
            else:
                brothers = self.GetBrotherAtoms(atom)
                if brothers:
                    if atom[0] != "H":
                        reference = atom
                    else:
                        for at in brothers:
                            if at[0] != "H":
                                reference = at
                                break
                        else:
                            reference = atom
                    self.DeltaFi[self.AtomNames.index(reference)] = 0
                    ref_fi = self.Diherdrals[self.AtomNames.index(reference)]
                    for at in brothers + [atom]:
                        if at == reference:
                            continue
                        else:
                            at_chi = self.Diherdrals[self.AtomNames.index(at)]
                            delta_fi = at_chi - ref_fi
                            self.DeltaFi[self.AtomNames.index(at)] = delta_fi
                            self.DeltaFiReference[self.AtomNames.index(at)] = reference
                else:
                    self.DeltaFi[index] = 0
        self.DeltaFi = tuple(self.DeltaFi)


class ROTAMERLIB:
    def __init__(self, aminoacid):
        self.Resname = ''
        self.NumberOfDihedrals = 0
        self.NumberOfPolarH = 0
        self.NumberOfDihedralAtoms = 0
        self.DihedralsDictio = {}
        self.PolarHydrogensDictio = {}
        self.ReadRotamerFile(aminoacid)

    def ReadRotamerFile(self, aminoacid):
        residue_rotamer_library_path = os.path.join(DIR, "Data/RotamerLibs/") + aminoacid.upper() + '.side'
        filein = open(residue_rotamer_library_path, 'r')
        file_text = filein.readlines()
        filein.close()
        rotamer_info = file_text[0].split()
        self.Resname = rotamer_info[1]
        self.NumberOfDihedralAtoms = int(rotamer_info[2])
        self.NumberOfDihedrals = int(rotamer_info[3])
        multiplier = float(rotamer_info[4])
        self.NumberOfPolarH = int(rotamer_info[5])
        ordered_keys = [element[:-1].strip('_') for element in file_text[1:self.NumberOfDihedralAtoms + 1]]
        self.DihedralsDictio = {key: tuple([float(x.strip().split()[y]) * multiplier
                                            for x in file_text[self.NumberOfDihedralAtoms + self.NumberOfPolarH + 1:]])
                                for key, y in zip(ordered_keys, range(len(ordered_keys)))}



from prody import AtomGroup

from PPP.mutational_module import AddAtoms, ModifyExistingAtoms
from PPP.program_own_classes import ZMATRIX
from PPP.coordinates_module import ComputeCartesianCoordinates, ComputeDihedral
from PPP.global_processes import FindInitialAndFinalResidues

__author__ = 'jelisa'


def DefineNewAtom(atom_name, element, coordinates, resname, resnum, chain_id):
    """
    This function creates a new AtomGroup instance containing one atom.
    """
    new_atom = AtomGroup()
    new_atom.setNames([atom_name])
    new_atom.setElements([element])
    new_atom.setCoords([coordinates])
    new_atom.setResnames(resname)
    new_atom.setResnums(resnum)
    new_atom.setChids(chain_id)
    new_atom.setAltlocs([''])
    new_atom.setBetas([0])
    new_atom.setIcodes([''])
    new_atom.setOccupancies([1])
    new_atom.setSegnames([''])
    new_atom.setSerials([0])
    # new_atom.setAnisous([[0.0, 0.0, 0.0]])
    return new_atom


def PlaceSpecialAtoms(old_residue, atom_name, structure, resnum, zmatrix, verbose=False):
    current_residue = structure.select("resnum {}".format(resnum)).copy()
    atom_position_in_zmatrix = zmatrix.AtomNames.index(atom_name)
    r = zmatrix.BondLengths[atom_position_in_zmatrix]
    deta = zmatrix.BondAngles[atom_position_in_zmatrix]
    element = "H"
    if atom_name == "H":
        previous_residue = structure.select("resnum {}".format(resnum - 1))
        if previous_residue is None:
            at3 = current_residue.select("name N")
            at2 = current_residue.select("name C")
            at1 = current_residue.select("name O")
            fi = 180
        else:
            at3 = current_residue.select("name N")
            at2 = previous_residue.select("name C")
            at1 = previous_residue.select("name O")
            fi = 180
    elif atom_name in ["HA2", "HA"]:
        # print '2'
        previous_residue = structure.select("resnum {}".format(resnum - 1))
        if previous_residue is None:
            at4 = current_residue.select("name N")
            at3 = current_residue.select("name CA")
            at2 = current_residue.select("name C")
            at1 = current_residue.select("name O")
            fi = -120 + ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0],
                                        at3.getCoords()[0], at4.getCoords()[0])
        else:
            at4 = current_residue.select("name C")
            at3 = current_residue.select("name CA")
            at2 = current_residue.select("name N")
            at1 = previous_residue.select("name C")
            fi = 120 + ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0],
                                       at3.getCoords()[0], at4.getCoords()[0])
    elif atom_name in ["HA3", "CB"]:
        previous_residue = structure.select("resnum {}".format(resnum - 1))
        if previous_residue is None:
            at4 = current_residue.select("name N")
            at3 = current_residue.select("name CA")
            at2 = current_residue.select("name C")
            at1 = current_residue.select("name O")
            fi = -240 + ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0],
                                        at3.getCoords()[0], at4.getCoords()[0])
        else:
            at4 = current_residue.select("name C")
            at3 = current_residue.select("name CA")
            at2 = current_residue.select("name N")
            at1 = previous_residue.select("name C")
            fi = 240 + ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0],
                                       at3.getCoords()[0], at4.getCoords()[0])
    elif atom_name == "OXT":
        at4 = current_residue.select("name O")
        at3 = current_residue.select("name C")
        at2 = current_residue.select("name CA")
        at1 = current_residue.select("name N")
        fi = 180 + ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0],
                                   at3.getCoords()[0], at4.getCoords()[0])
        element = "O"
    elif atom_name in ["H1", "H2", "H3"]:
        at3 = current_residue.select("name N")
        at2 = current_residue.select("name CA")
        if current_residue.getResnames()[0] == "GLY":
            at1 = current_residue.select("name HA2")
        else:
            at1 = current_residue.select("name CB")
        # This magic numbers come from placing the hydrogens in maestro
        # and computing the angle and the dihedrals.
        fi = 60 + 120 * (int(atom_name[1]) - 1)  # dihedral between C-Calpha-N-hydrogens in maestro
        deta = 109.53580  # Angle between Calpha-N-Hydrogens in maestro.
    else:
        return old_residue
    coords = ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi)
    new_atom = DefineNewAtom(atom_name, element, coords, current_residue.getResnames()[:1],
                             current_residue.getResnums()[:1], current_residue.getChids()[:1])
    # print len(old_residue.getAnisous()), len(new_atom.getAnisous())
    residue = old_residue + new_atom
    return residue


def PlaceHydrogen(old_residue, atom_name, zmatrix):
    at1_name, at2_name, at3_name, r, deta, zmatrix_fi = zmatrix.GetData2ComputeCoords(atom_name)
    # print 'a', old_residue.getNames(), old_residue.getResnames()
    # print 'b', zmatrix.AtomNames
    at1 = old_residue.select("name {}".format(at1_name))
    at2 = old_residue.select("name {}".format(at2_name))
    at3 = old_residue.select("name {}".format(at3_name))
    delta_fi = zmatrix.DeltaFi[zmatrix.AtomNames.index(atom_name)]
    reference_atom_name = zmatrix.DeltaFiReference[zmatrix.AtomNames.index(atom_name)]
    reference_atom = old_residue.select("name {}".format(reference_atom_name))
    # print 'd', zmatrix.DeltaFi
    # print 'c', at1, at2, at3, reference_atom_name, atom_name
    fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0],
                         reference_atom.getCoords()[0]) + delta_fi
    coords = ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi)
    new_atom = DefineNewAtom(atom_name, "H", coords, old_residue.getResnames()[:1],
                             old_residue.getResnums()[:1], old_residue.getChids()[:1])
    # print old_residue.getAnisous(), new_atom.getAnisou()
    new_residue = old_residue + new_atom
    return new_residue


def FixStructure(initial_structure, residues2fix, gaps, charge_terminals, debug=False):
    final_structure = None
    for chain in initial_structure.iterChains():
        if chain.getChid() in gaps.keys():
            gaps_e = [x[0] for x in gaps[chain.getChid()]]
            gaps_b = [x[1] for x in gaps[chain.getChid()]]
        else:
            gaps_e = []
            gaps_b = []
        chain_id = chain.getChid()
        if chain_id == " ":
            print("#####################")
            print("WARNING: There is some chain id missing!!!!!!!")
            print("#####################")
        current_structure = initial_structure.select("chain `{}`".format(chain_id)).copy()
        new_chain = None
        initial_residue, final_residue = FindInitialAndFinalResidues(current_structure)
        for residue in current_structure.iterResidues():
            res_id = " ".join([residue.getResname(), str(residue.getResnum()), residue.getChid()])
            resnum = residue.getResnum()
            resname = residue.getResname()
            old_res = current_structure.select("resname `{}` and resnum `{}`".format(residue.getResname(),
                                                                                     residue.getResnum())).copy()
            if res_id not in residues2fix.keys():
                new_residue = old_res
            else:
                if charge_terminals:
                    if resnum == initial_residue or resnum in gaps_b:
                        zmatrix = ZMATRIX(resname + 'B')
                        atoms2add = {"H1", "H2", "H3"}.union(residues2fix[res_id]['add']).difference(
                            set(residue.getNames()))
                    elif resnum == final_residue or resnum in gaps_e:
                        zmatrix = ZMATRIX(resname + 'E')
                        atoms2add = {"OXT"}.union(residues2fix[res_id]['add']).difference(set(residue.getNames()))
                    else:
                        zmatrix = ZMATRIX(resname)
                        atoms2add = residues2fix[res_id]['add']
                else:
                    zmatrix = ZMATRIX(resname)
                    atoms2add = residues2fix[res_id]['add']
                if residues2fix[res_id]['delete']:
                    print("  * Removing the charge from the terminal residue {0:3} {1:1} {2:3}. " \
                          "The removed atoms are : {3}".format(residue.getResname(),
                                                               residue.getChid(), residue.getResnum(),
                                                               " ".join(residues2fix[res_id]['delete'])))
                    new_residue = residue.select('not name {}'.format(" ".join(residues2fix[res_id]['delete']))).copy()
                    old_res = new_residue
                if residues2fix[res_id]['modify']:
                    atomnames_of_2_letters = ["FE"]
                    print("  * Modifying the residue {0:3} {1:1} {2:3}:".format(residue.getResname(),
                                                                                residue.getChid(),
                                                                                residue.getResnum(),
                                        " ".join(residues2fix[res_id]['add'])))
                    new_residue = ModifyExistingAtoms(old_res, residues2fix[res_id]['modify'],
                                                      atomnames_of_2_letters, 0, 1, zmatrix)
                    old_res = new_residue
                if residues2fix[res_id]['add']:
                    print("  * Adding to the residue {0:3} {1:1} {2:3} the atoms:" \
                          " {3}".format(residue.getResname(), residue.getChid(), residue.getResnum(),
                                        " ".join(residues2fix[res_id]['add'])))
                    # print " Adding to the residue '{}' the following atoms\n  {}".format(res_id, residues2fix[res_id])
                    if zmatrix.Name is None:
                        print("  ** The residue {} {} doesn't have a template, so it cannot be fixed.\n" \
                              "  ** PELE won't work! Check it!".format(residue.getResname(), residue.getResnum()))
                        continue
                    if debug:
                        print("ZMATRIX used:", zmatrix.Name)
                        print('atoms to add:', atoms2add)
                        print('residues2fix:', residues2fix[res_id])
                    residue_info = {"fin_resname": resname, "resnum": resnum, "chain": residue.getChid()}
                    if resname == "HOH":
                        print("  ** The water {0:3} {1:1} {2:3} is missing the atoms: {3}.\n" \
                              "  ** The program won't place them. \n" \
                              "  ** PELE will crash".format(resname, residue.getChid(), resnum,
                                                            ", ".join(residues2fix[res_id]['add'])))
                        new_residue = current_structure.select('resnum {}'.format(resnum)).copy()
                    else:
                        try:
                            zmatrix.ComputeDeltaFi()
                        except ValueError:
                            print("Something went wrong computing the delta phi for the residue {} {}.\n" \
                                  "Check the template ZMATRIX".format(resname, resnum))
                        for name in zmatrix.AtomNames:
                            if name in atoms2add:
                                if name in ["H", "HA2", "HA", "HA3", "CB", "OXT", "H1", "H2", "H3"]:
                                    if debug:
                                        new_residue = PlaceSpecialAtoms(old_res, name, current_structure,
                                                                        residue.getResnum(), zmatrix, True)
                                    else:
                                        new_residue = PlaceSpecialAtoms(old_res, name, current_structure,
                                                                        residue.getResnum(), zmatrix)
                                    old_res = new_residue
                                elif name[0] == "H" and zmatrix.DeltaFi[zmatrix.AtomNames.index(name)] != 0:
                                    new_residue = PlaceHydrogen(old_res, name, zmatrix)
                                    old_res = new_residue
                                else:
                                    new_residue = AddAtoms(old_res, [name], ["FE"], residue_info, zmatrix, False)
                                    old_res = new_residue
            if new_residue.getAnisous() is not None and not len(new_residue.getAnisous()) == new_residue.numAtoms():
                """If this values don't match this will give problems later when copying the structure.
                The magic number 6 used to define the new_anisous is due to the fact that the anisotropic
                arrays have six positions."""
                new_anisous = residue.getAnisous()
                new_anisous.resize(new_residue.numAtoms(), 6)
                new_residue.setAnisous(new_anisous)
                if debug:
                    print("checking the concordance between the anisou and the number of atoms")
                    print("len anisous: {}, number of atoms: {}".format(len(new_residue.getAnisous()),
                                                                        new_residue.numAtoms()))
            if new_chain is None:
                new_chain = new_residue
            else:
                new_chain = new_chain + new_residue
        if final_structure is None:
            final_structure = new_chain
        else:
            final_structure = final_structure + new_chain
    final_structure.setTitle("Structure with Hydrogens")
    return final_structure



import sys

from numpy import asarray, vstack, cos, cross, dot, pi, sin, transpose, linalg, arctan2


__author__ = 'jelisa'


fixed_atoms = ["N", "H", "CA", "HA", "C", "O", "CB"]


def Grades2Radians(angle):
    """This is a small function to convert from degrees to radians a given angle.
    :rtype : a float with the angle in radians
    :param angle: an angle in degrees
    :return: the angle in radians
    """
    return (angle / 180.) * pi


def Radians2Degrees(angle):
    """
    :rtype : float
    :param angle: float with the angle in radians
    :return: the angle in degrees
    """
    return (angle * 180) / pi


def ComputeAngleDifference(angle1, angle2):
    """
    This function computes the difference between two angles.
    :rtype : float
    :param angle1: float with the first angle
    :param angle2: float with the second angle
    :return: 
    """
    if angle1 < 0:
        inicial = angle1 + 360
    else:
        inicial = angle1
    if angle2 < 0:
        final = angle2 + 360
    else:
        final = angle2
    diff = inicial - final
    return diff


# noinspection PyTypeChecker
def ComputeDihedral(at1, at2, at3, at4, v=False):
    """
    This function implements the calculation of the Dihedral using the arctan as explained
    in the link: http://math.stackexchange.com/questions/47059/how-do-i-calculate-a-dihedral-angle-given-cartesian-coordinates
    date 14/05/2015. It's the same algorithm as the one implemented in PELE-1.5++ at date 15/05/2015
    :rtype : float
    :param at1: coordinates of the first atom
    :param at2: coordinates of the second atom
    :param at3: coordinates of the third atom
    :param at4: coordinates of the fourth atom
    :return: the dihedral angle in degrees between the four atoms.
    """
    b1 = at1 - at2
    b2 = at2 - at3
    b3 = at3 - at4
    c_b1b2 = cross(b1, b2)
    c_b2b3 = cross(b2, b3)
    u_b2 = b2 / linalg.norm(b2)
    # print u_b2
    n1 = c_b1b2 / linalg.norm(c_b1b2)
    n2 = c_b2b3 / linalg.norm(c_b2b3)
    # print "compute dihedrals, a", n1.shape, n2.shape
    m1 = cross(n1, u_b2)
    x = dot(n1, n2)
    y = dot(m1, n2)
    z = dot(u_b2, n2)
    # print z
    if z > 0.0001:
        print("Z should be almost 0, something in wrong!")
        sys.exit()
    dihedral = arctan2(y, x)
    if v:
        print("dihedral in rad:", dihedral)
    return Radians2Degrees(dihedral)


# noinspection PyTypeChecker
def ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi):
    """
    This function implements the SN-NeRF algorithm described in Parson et al., 2005  (Ref. 14)
    :rtype : array with the coordinates
    :param at1: An AtomGroup from prody containing the atom 1 (atom A in the article)
    :param at2: An AtomGroup from prody containing the atom 2 (atom B in the article)
    :param at3: An AtomGroup from prody containing the atom 3 (atom C in the article)
    :param r: The distance between the atom to place and at3.
    :param deta: The angle defined by at2 and the atom to place with the at3 as vertex
    :param fi: The dihedral angle defined by at1, at2, at3 and the atom to place
    """

    rad_deta = Grades2Radians(deta)
    rad_fi = Grades2Radians(fi)
    # print 'compute coords', fi, rad_fi
    bc = at3.getCoords()[0] - at2.getCoords()[0]
    ab = at2.getCoords()[0] - at1.getCoords()[0]
    u_bc = bc / linalg.norm(bc)
    n = cross(ab, u_bc) / linalg.norm(cross(ab, u_bc))
    m = transpose(vstack([u_bc, cross(n, u_bc), n]))
    d2 = asarray([-r * cos(rad_deta), r * cos(rad_fi) * sin(rad_deta), r * sin(rad_fi) * sin(rad_deta)])
    coords = dot(m, d2) + at3.getCoords()[0]
    return asarray(coords)


def GenerateCoordinatesFromZmatrix(initial_residue, atoms2add, zmatrix):
    atoms2add_coords = []
    for atom in atoms2add:
        at1_name, at2_name, at3_name, r, deta, fi = zmatrix.GetData2ComputeCoords(atom)
        # print atom, at1_name, at2_name, at3_name
        # print zmatrix.Name, zmatrix.AtomNames
        at3 = initial_residue.select('name {}'.format(at3_name))
        at2 = initial_residue.select('name {}'.format(at2_name))
        at1 = initial_residue.select('name {}'.format(at1_name))
        # print initial_residue.getNames()
        # print at1, at2, at3
        atoms2add_coords.append(ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi))
    return asarray(atoms2add_coords)


def ChangeResidueCoordinates(initial_residue, zmatrix, rotamer_library, dihedral_index):
    """
    This function changes the coordinates of a residue according to the
    information in the zmatrix and the rotamer library used in PELE.

    :type rotamer_library: ROTAMERLIB
    :type zmatrix: ZMATRIX
    :type initial_residue: AtomGroup of prody
    :type dihedral_index: int
    :rtype : AtomGroup of prody
    :param initial_residue: atom group of prody with the initial residue to modify
    :param zmatrix: ZMATRIX object with the information for the residue
    :param rotamer_library: ROTAMERLIB object containing the information of the rotamer library for the residue
    :param dihedral_index: the index that indicates which rotamer should be used
    :return: AtomGroup of prody with the coordinates changed
    """
    residue = initial_residue.copy()
    dihedral_atoms_brothers = []
    for atom_name in rotamer_library.DihedralsDictio.keys():
        dihedral_atoms_brothers.extend(zmatrix.GetBrotherAtoms(atom_name))
    atoms_yet_to_place = []
    for atom_name in zmatrix.AtomNames:
        if atom_name in fixed_atoms:
            continue
        elif atom_name[0] == "H":
            atoms_yet_to_place.append(atom_name)
            continue
        else:
            at1_name, at2_name, at3_name, r, deta, zmatrix_fi = zmatrix.GetData2ComputeCoords(atom_name)
            at3 = residue.select("name {}".format(at3_name))
            at2 = residue.select("name {}".format(at2_name))
            at1 = residue.select("name {}".format(at1_name))
            atom = residue.select("name {}".format(atom_name))
            delta_fi = zmatrix.DeltaFi[zmatrix.AtomNames.index(atom_name)]
            if atom_name in rotamer_library.DihedralsDictio.keys():
                initial_fi = rotamer_library.DihedralsDictio[atom_name][dihedral_index]
            elif atom_name in dihedral_atoms_brothers:
                dihedral_brother = [at for at in zmatrix.GetBrotherAtoms(atom_name)
                                    if at in rotamer_library.DihedralsDictio.keys()]
                try:
                    dihedral_brother[0]
                except IndexError:
                    print("Something went wrong when checking brothers while moving the residue to solve clashes.")
                    sys.exit()
                else:
                    initial_fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0],
                                                 residue.select("name {}".format(dihedral_brother[0])).getCoords()[0])
            elif zmatrix.DeltaFi[zmatrix.AtomNames.index(atom_name)] != 0 and zmatrix.GetBrotherAtoms(atom_name):
                brother = zmatrix.GetBrotherAtoms(atom_name)[0]
                initial_fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0],
                                             residue.select("name {}".format(brother)).getCoords()[0])
            else:
                initial_fi = zmatrix_fi

            fi = initial_fi + delta_fi
            coords = ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi)
            atom.setCoords(coords)
    for atom_name in atoms_yet_to_place:
        at1_name, at2_name, at3_name, r, deta, zmatrix_fi = zmatrix.GetData2ComputeCoords(atom_name)
        at3 = residue.select('name {}'.format(at3_name))
        at2 = residue.select('name {}'.format(at2_name))
        at1 = residue.select('name {}'.format(at1_name))
        atom = residue.select('name {}'.format(atom_name))
        delta_fi = zmatrix.DeltaFi[zmatrix.AtomNames.index(atom_name)]
        if atom_name in dihedral_atoms_brothers:
            dihedral_brother = [at for at in zmatrix.GetBrotherAtoms(atom_name)
                                if at in rotamer_library.DihedralsDictio.keys()]
            try:
                dihedral_brother[0]
            except IndexError:
                print("Something went wrong when checking brothers while moving the residue to solve clashes.")
                sys.exit()
            else:
                initial_fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0],
                                             residue.select('name {}'.format(dihedral_brother[0])).getCoords()[0])
        elif zmatrix.DeltaFi[zmatrix.AtomNames.index(atom_name)] != 0 and zmatrix.GetBrotherAtoms(atom_name):
            brother = zmatrix.GetBrotherAtoms(atom_name)[0]
            initial_fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0],
                                         residue.select("name {}".format(brother)).getCoords()[0])
        else:
            initial_fi = zmatrix_fi
        fi = initial_fi + delta_fi
        coords = ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi)
        atom.setCoords(coords)
    return residue


def ModifyCoords(initial_residue, zmatrix, atom_name):
    zmatrix.ComputeDeltaFi()
    at1 = initial_residue.select("name N")
    at2 = initial_residue.select("name C")
    at3 = initial_residue.select("name CA")
    atom = initial_residue.select("name {}".format(atom_name))
    r = zmatrix.BondLengths[zmatrix.AtomNames.index(atom_name)]
    deta = zmatrix.BondAngles[zmatrix.AtomNames.index(atom_name)]
    fi = ComputeDihedral(at1.getCoords()[0], at2.getCoords()[0], at3.getCoords()[0], atom.getCoords()[0])
    coords = ComputeCartesianCoordinates(at1, at2, at3, r, deta, fi)
    return coords


def ModifyCoordinatesPRO(residue, zmatrix, atom_name, reference_atom):
    parent_name = zmatrix.GetParentName(atom_name)
    parent_atom = residue.select("name {}".format(parent_name)).copy()
    atom = residue.select("name {}".format(reference_atom)).copy()
    parent_atom_coords = parent_atom.getCoords()
    vector = atom.getCoords() - parent_atom_coords
    unitary_vector = vector / linalg.norm(vector)
    r = zmatrix.BondLengths[zmatrix.AtomNames.index(atom_name)]
    atom_coords = parent_atom_coords + unitary_vector * r
    return atom_coords



import numpy as np

CONSTRAINT = '{{ "type": "constrainAtomToPosition", "springConstant": 10, "equilibriumDistance": 0.0, "constrainThisAtom": "{}:{}:{}" }},'

def search_smart_pattern(ligand, smiles):
    from rdkit import Chem
    # Get smiles pattern
    patt = Chem.MolFromSmarts(smiles)
    Chem.Kekulize(patt)
    # Get ligand
    m = Chem.MolFromPDBFile(ligand)
    Chem.Kekulize(m)
    # Substructure search
    atom_idxs = np.array(m.GetSubstructMatch(patt))
    try:
        atoms_selected = np.array(m.GetAtoms())[atom_idxs]
    except IndexError:
        raise ValueError("No substructure was recognised with {}  as the smile pattern and {} as the molecule".format(
        smiles, Chem.MolToSmiles(m)))
    # Retrieve atom names and residue names of the common atoms
    atom_names = [atom.GetMonomerInfo().GetName().replace(" ", "_") for atom in atoms_selected]
    chain_names = [atom.GetPDBResidueInfo().GetChainId() for atom in atoms_selected]
    residue_numbers = [atom.GetPDBResidueInfo().GetResidueNumber() for atom in atoms_selected]
    # Get coords
    conformer = m.GetConformer()
    coords = np.array(conformer.GetPositions())[atom_idxs]
    return atom_names, chain_names, coords, residue_numbers

def extract_constraint_from_smiles(ligand, smiles):
    from rdkit import Chem
    m = Chem.MolFromPDBFile(ligand)
    atom_names, chain_names, coords, residue_numbers = search_smart_pattern(ligand, smiles)
    constants = [CONSTRAINT.format(chain_name, resname, atom_name) for atom_name, chain_name, coord, resname in zip(atom_names, chain_names, coords, residue_numbers)]
    return constants



if __name__ == "__main__":
    ligand = "ligand.pdb"
    smiles = "CCC([O-])O"
    constraints = extract_constraint_from_smiles(ligand, smiles)



__author__ = 'jelisa'

program_description = "This program will substitute plop and ppp for PELE. It'll check the structure\n" \
                      " the atom names in the future, for now it does mutations and checks-formats the\n" \
                      " names of the residues, produces a .pdb file that PELE can read without problems."
ipdb_description = "This parameter is mandatory and specifies the path to\n" \
                   " the .pdb file with the structure to mutate.\n" \
                   "This structure should have all the atoms present,\n" \
                   " including the hydrogens."
opdb_description = "This specifies the path to write the final pdb.\n" \
                   "If it isn't specified the program will generate\n" \
                   " a pdb with the same name as the initial pdb but\n" \
                   " with the suffix _processed."
mutation_description = "This parameter specifies the desired mutation(s).\n" \
                       "It should be a string wiht the desired mutation(s)\n" \
                       " in the following format:\n" \
                       "  'residue XXX N to YYY'\n" \
                       "  where:\n" \
                       "    - XXX: the original residue\n" \
                       "    - N: the number of the residue\n" \
                       "    - YYY: the desired aminoacid.\n" \
                       "The names of the residues should be the three letters\n" \
                       " code for the aminoacid.\n" \
                       "EXCEPTIONS: the hystidine you should use one of the \n" \
                       " following names: HID, HIP, HIE"
ff_description = "This specifies the force-field to use. The supported force-fields are:\n" \
                 " - OPLS2005 (default)\n" \
                 " - OPLS2001\n" \
                 " - AMBER99sbBSC0\n"
complete_tests = "When this option is choosen the program will undergo\n" \
                 " all the tests implemented in the tests modules."
mutants_from_file_description = "This option specifies a path to read the output names and the\n" \
                                " mutations from a file.\n" \
                                "The file should have one mutation per line and no blank lines.\n" \
                                "Each line should have the following format:\n" \
                                "  output_name-->XXX N C YYY\n" \
                                "    --> indicates a tab space\n" \
                                "    XXX is the initial aminoacid in 3 letters code\n" \
                                "    N the residue number\n" \
                                "    C the chain to modify\n" \
                                "    YYY the desired aminoacid in 3 letters code\n"
mutant_multiple_description = "This parameter creates one structure with all the\n" \
                              " mutations specified. But it doesn't check for clashes."
charge_terminals_description = "When this option is present the program will charge all the terminals residues," \
                               "including those in gaps."
remove_terminal_missing_description = "Whe this option is chosen the program will remove the terminal residues " \
                                      "in case they are missing heavy atoms in the backbone of the protein" \
                                      "(that is: N, CA, C, O)."
no_gaps_ter_description = "When this option is present the program won't add a TER mark whenever it finds " \
                       "a gap in the sequence."
pdb_resolution_description = "A float indicating the pdb resolution, this number will be used as the maximum allowed " \
                             "between the N atom of a residue and the C atom of the previous residue when checking" \
                             "for gaps. In this check the minimum distance that will be used is 1.55A."
make_unique_description = "This option is optional and should be present if the user wants to make the names " \
                          "in a given CHAIN unique."
remove_terminal_missing = "When this option is present the program will eliminate the first or last residue if they " \
                          "have missing atoms, and fix the next or previous residues to be the terminal one."


import argparse
import os
import sys
from multiprocessing import Pool, cpu_count
import subprocess
import PPP.mut_prep4pele as mut
import PPP.enviroment_parameters as env
import Utilities.template_builder as tb
import Utilities.helpers as hp
import Utilities.best_structs as bs
import shutil
import AdaptivePELE.adaptiveSampling as ad

PELE_BIN = os.path.join(env.pele_folder_path, "bin/Pele_mpi")
DIR = os.path.dirname(os.path.realpath(__file__))
ADAPTIVE_CONTROL_FILE = os.path.join(DIR, "Template/adaptive.conf")
PELE_CONTROL_FILE = os.path.join(DIR, "Template/pele.conf")
CMD = "python -m AdaptivePELE.adaptiveSampling {}"


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_system", help="Input to apply the mutations to")
    parser.add_argument("file", help="File to read mutations from. Each line needs to follow the next format:  mut1.pdb ARG 164 ALA", type=str)
    parser.add_argument("--box", nargs="+", help="Center of the box")
    parser.add_argument("--cpus", type=int, help="Cpus to use", default=1)
    args = parser.parse_args()
    return args.input_system, args.file, args.box, args.cpus

def main(system, mutations_file, box, cpus):
    mutations = retrieve_mutation(mutations_file)
    for mutants in mutations:
        mutants_for_pele = []
        for mutation_info in mutants:
            print(mutation_info.split(" ",1))
            output, mutation = mutation_info.split(" ",1)
            ini_res, resnum, chain, end_res = mutation.split()
            mutation = {'fin_resname': end_res, 'resnum': resnum, 'chain': "", 'ini_resname': ini_res}
            mutants_for_pele.append(mut.main(system, 2.5, mutation=[mutation], output_pdb=[output])[0])
        folders = [ os.path.abspath(os.path.splitext(mutant)[0]) for mutant in mutants_for_pele ]
        pool = Pool(cpus)
        pool.map(launch_mutation_PELE, mutants_for_pele)
        pool.close()
        system = bs.main(path=folders, criteria="Binding Energy", n_structs=1, sort_order="min", out_freq=1, output="Mutation")




def retrieve_mutation(mutations_file):
    with open(mutations_file, "r") as f:
        line = f.readline()
        mutations = []
        mutation = []
        while line:
            if "MUTATION" not in line:
                if line.strip():
                    mutation.append(line.strip("\n").strip().strip("\t"))
            else:
                if mutation:
                    mutations.append(mutation)
                mutation = []
            line = f.readline()
        mutations.append(mutation)
    return mutations



def launch_mutation_PELE(mutant):
    root_name = os.path.abspath(os.path.splitext(mutant)[0])
    ad_conf = prepare_control_file(root_name, mutant, box, adaptive=True)
    _ = prepare_control_file(root_name, mutant, box, adaptive=False)
    subprocess.call(CMD.format(ad_conf).split())

def prepare_control_file(root_name, mutant, box, adaptive=False):
    conf_name = root_name + "_adaptive.conf" if adaptive else root_name + "_pele.conf"
    KEYWORDS = {"OUTPUT": root_name, "STRUCT": os.path.abspath(mutant), "PELE": root_name + "_pele.conf"} if adaptive else {"BOX" : ",".join(box)}
    template_file = ADAPTIVE_CONTROL_FILE if adaptive else  PELE_CONTROL_FILE
    shutil.copy(template_file, conf_name)
    tb.TemplateBuilder(conf_name, KEYWORDS)
    return conf_name


if __name__ == "__main__":
    system, mutations, box, cpus = parse_args()
    main(system, mutations, box, cpus)



import PPP
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from distutils.extension import Extension

here = path.abspath(path.dirname(__file__))
ext_modules = []
cmdclass = {}
# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
     long_description = f.read()

setup(
    name="PPPele",
    version=PPP.__version__,
    description='Protein PELE preparation',
    long_description=long_description,
    url="https://github.com/NostrumBioDiscovery/PPP",
    author='Daniel Soler',
    author_email='daniel.soler@nostrumbiodiscovery.com',
    packages=find_packages(exclude=['docs', 'tests']),
    package_data={"PPP/Data/Templates/OPLS2005/HeteroAtoms/": ['*'],
                  "PPP/Data/Templates/OPLS2005/Protein": ['*'],
                  "PPP/Data/RotamerLibs/": ['*']},
    include_package_data=True,
    install_requires=["biopython", "prody", "pytest", "scipy"],
    cmdclass=cmdclass,
    ext_modules=ext_modules  # accepts a glob pattern
)



import copy

from prody import calcDistance

from PPP.checks_module import CheckClashes
from PPP.coordinates_module import ChangeResidueCoordinates
from PPP.global_processes import FindInitialAndFinalResidues
from PPP.global_variables import protein_atomnames_dictionary, supported_aminoacids, supported_metals
from PPP.program_own_classes import ROTAMERLIB

__author__ = 'jelisa'


def SolveClashes(initial_structure, initial_clashes, mutation, zmatrix, initial_residue_number, final_residue_number):
    print("Trying to solve the clashes.")
    dihedral2check = 0
    initial_part_of_the_protein = initial_structure.select("resnum < {}".format(mutation["resnum"])).copy()
    final_part_of_the_protein = initial_structure.select("resnum > {}".format(mutation["resnum"])).copy()
    initial_residue = initial_structure.select("resnum {}".format(mutation["resnum"])).copy()
    resname = initial_residue.getResnames()[0]

    if resname in ["HIE", "HID", "HIP"]:
        resname = "HIS"
    elif resname in ["LYN"]:
        resname = "LYS"
    residue_rotamer_lib = ROTAMERLIB(resname)
    maximum_number_of_trials = residue_rotamer_lib.NumberOfDihedrals
    initial_number_of_clashes = 0
    for values in initial_clashes.values():
        initial_number_of_clashes += len(values)
    print("Starting number of heavy atoms clashes: {}".format(len(initial_clashes.keys())))
    print("Starting number of total clashes: {}".format(initial_number_of_clashes))

    zmatrix.ComputeDeltaFi()
    while maximum_number_of_trials > 0:
        maximum_number_of_trials -= 1
        new_residue = ChangeResidueCoordinates(initial_residue.copy(), zmatrix, residue_rotamer_lib, dihedral2check)
        if mutation["fin_resname"] == "PRO":
            CD_N_distance = calcDistance(new_residue.select("name N"), new_residue.select("name CD"))
            if CD_N_distance > 1.5:
                dihedral2check += 1
                continue
        dihedral2check += 1
        mutated_structure = initial_part_of_the_protein + new_residue + final_part_of_the_protein
        new_clashes = CheckClashes(mutated_structure, mutation, zmatrix, initial_residue_number, final_residue_number)
        current_number_of_clashes = 0
        for values in new_clashes.values():
            current_number_of_clashes += len(values)
        if new_clashes == {}:
            initial_residue = new_residue
            break
        elif len(new_clashes.keys()) < len(initial_clashes.keys()):

            print('The number of clashing heavy atoms in the residue now is: {} and clashes with {} atoms'.format(
                len(new_clashes.keys()),
                current_number_of_clashes))
            initial_residue = new_residue.copy()
            initial_number_of_clashes = current_number_of_clashes
            initial_clashes = new_clashes
        elif len(new_clashes.keys()) == len(initial_clashes.keys()):
            if current_number_of_clashes < initial_number_of_clashes:
                print("The number of clashing heavy atoms is still the same but now they clash with {} atoms.".format(
                    current_number_of_clashes))
                initial_residue = new_residue.copy()
                initial_number_of_clashes = current_number_of_clashes
    return initial_part_of_the_protein + initial_residue + final_part_of_the_protein


def FixAtomNames(initial_structure, gaps={}, no_gaps={}, debug=False):

    correct_structure = initial_structure.copy()

    for chain in correct_structure.iterChains():
        # for chain in structure_without_waters.iterChains():
        if chain.getChid() in gaps.keys():
            gaps_residues = [y for x in gaps[chain.getChid()] for y in x]
            gaps_e = [x[0] for x in gaps[chain.getChid()]]
            gaps_b = [x[1] for x in gaps[chain.getChid()]]
        else:
            gaps_residues = []
            gaps_b = []
            gaps_e = []
        if chain.getChid() in no_gaps.keys():
            no_gaps_residues = [y for x in no_gaps[chain.getChid()] for y in x]
        else:
            no_gaps_residues = []
        initial_res, final_res = FindInitialAndFinalResidues(chain)
        if debug:
            print("working with chain {}".format(chain.getChid()))
        for residue in chain.iterResidues():
            tmp_dictio = copy.deepcopy(protein_atomnames_dictionary)
            resname = residue.getResname()
            try:
                possible_names = tmp_dictio[resname]
            except KeyError:
                print('   * The residue {} is not an aa nor a water.'.format(resname))
                possible_names = [["CL"], ["CU"], ["FE1"], ["FE2"], ["ZN"], ["MG"]]
                heteroatom = True
            else:
                if resname == 'HOH':
                    heteroatom = True
                else:
                    heteroatom = False
            if residue.getResnum() == initial_res or residue.getResnum() in gaps_b:
                # This magic number 0 comes from the fact that for all the aminoacids
                # except the proline the first set of atoms in the protein_atomnames_dictionary
                # should have the hydrogen H from the nitrogen, but if the residue is the
                # first one in the protein it should be changed into H3 which belongs to
                # the "END" keyword in the names dictionary.
                if 'H' in possible_names[0]:
                    possible_names.pop(0)
            if residue.getResnum() in [initial_res, final_res] + gaps_residues:
                # The other in this step is really important it should always be the
                # possible and then the ending possibilities, otherwise it will create
                # problems for the ending residues.
                possible_names = possible_names + tmp_dictio["END"]

            # if debug and residue.getResnum() == debug[0] and chain.getChid() == debug[1]:
            #     print residue.getResname(), residue.getResnum(), residue.getChid(), residue.getNames()
            for atom in residue.iterAtoms():
                atom_found = False
                atom_name = atom.getName()
                # if debug and residue.getResnum() == debug[0] and chain.getChid() == debug[1]:
                #     print 'pos', atom_name, possible_names
                for atoms in possible_names:
                    if atom_name in atoms:
                        old_atom_name = atom_name
                        atom_name = atoms[0]
                        atom_found = True
                        possible_names.pop(possible_names.index(atoms))

                        break
                if not atom_found and not heteroatom:
                    if atom_name in ['HXT', 'H1', "H2"]:
                        if residue.getResnum() in gaps_residues or residue.getResnum() in [initial_res, final_res]:
                            pass
                        elif residue.getResnum() in no_gaps_residues:
                            print("   * The residue {} won't be considered as a gap, if it really is one," \
                                  " let the developer know".format("{2} {1} {0}".format(residue.getResnum(),
                                                                                        residue.getChid(),
                                                                                        residue.getResname())))
                    else:
                        print("   * The Atom {} from residue {} {} {} doesn't have a valid name.".format(atom_name,
                                                                                                         resname,
                                                                                                         atom.getChid(),
                                                                                                         atom.getResnum()))
                atom.setName(atom_name)
                # if debug: break
    return correct_structure


def WritingAtomNames(initial_structure):
    correct_names_structure = initial_structure.copy()
    for atom in correct_names_structure.iterAtoms():
        atom_name = atom.getName()
        if atom.getResname() in supported_metals:
            atom_name = "{0}  ".format(atom_name)
        elif atom_name == "1HW":
            atom_name = "1HW "
        elif atom_name == "2HW":
            atom_name = "2HW "
        elif atom_name == "FE1":
            atom_name = "FE1 "
        elif atom_name == "FE2":
            atom_name = "FE2 "
        elif atom.getResname() == "NA":
            atom_name = "NA  "
        elif atom_name == "CL":
            atom_name = "CL  ".format(atom_name)
        elif len(atom_name) == 1:
            atom_name = " " + atom_name + "  "
        elif len(atom_name) == 2:
            atom_name = " " + atom_name + " "
        elif len(atom_name) == 3:
            atom_name = " " + atom_name
        atom.setName(atom_name)
    return correct_names_structure


def FixStructureResnames(initial_structure, ligand_chain=False):
    """
    This function changes the residues names so they match with the names required by PELE.
    """
    structure = initial_structure.copy()
    for res in structure.iterResidues():
        resname = res.getResname()
        ligand_structure = None
        # print ligand_chain == res.getChid(), resname in supported_aminoacids
        if ligand_chain == res.getChid() and resname in supported_aminoacids:
            ligand_structure = structure.chain_Z.copy()
        if ligand_structure is not None and ligand_structure.hetero is not None:
            print("INFO: Renaming the ligand to LIG in structure {}".format(initial_structure.getTitle()))
            resname = "LIG"
            res.setResnum(1)
        if resname == "HIS":
            if res.numAtoms() in [10, 16]:
                resname = "HIE"
            elif res.numAtoms() == 19:
                if ("HE2" in res.getNames() and "HD1" in res.getNames()) or\
                        ("HNE" in res.getNames() and "HND" in res.getNames()):
                    resname = "HIP"
                elif "HE2" in res.getNames() or "HNE" in res.getNames():
                    resname = "HIE"
                elif "HD1" in res.getNames() or "HND" in res.getNames():
                    resname = "HID"
            elif res.numAtoms() == 18:
                if "HE2" in res.getNames() and "HD1" in res.getNames() or\
                        ("HNE" in res.getNames() and "HND" in res.getNames()):
                    resname = "HIP"
                elif "HE2" in res.getNames() or "HNE" in res.getNames():
                    resname = "HIE"
                elif "HD1" in res.getNames() or "HND" in res.getNames():
                    resname = "HID"
                else:
                    print("Make sure the histidine is correctly protonated and then, only then, explain the bug" \
                          " to the the developer.")
            elif res.numAtoms() == 17:
                if "HE2" in res.getNames() or "HNE" in res.getNames():
                    resname = "HIE"
                elif "HD1" in res.getNames() or "HND" in res.getNames():
                    resname = "HID"
            else:
                print("  * WARNING: The hystidine {} has an incorrect number " \
                      "of atoms. PELE won't work correctly".format("{} {}".format(res.getResnum(), res.getChid(),)))
        elif resname == "LYS":
            if res.numAtoms() == 21:
                resname = "LYN"
            elif "OXT" in res.getNames() and res.numAtoms() == 22:
                resname = "LYN"
            elif "H1" in res.getNames() and res.numAtoms() == 23 and "HZ3" not in res.getNames():
                    resname = "LYN"
        elif resname == "GLU":
            if res.hydrogen is not None:
                if res.numAtoms() == 16 and res.hydrogen.numAtoms() == 7 and "H1" not in res.getNames() and \
                                "HXT" not in res.getNames():
                    resname = "GLH"
                elif res.numAtoms() == 18 and res.hydrogen.numAtoms() == 9:
                    resname = "GLH"
                elif res.numAtoms() == 17 and res.oxygen.numAtoms() == 4:
                    resname = "GLH"
        elif resname == "ASP":
            if res.numAtoms() == 13 and "OXT" not in res.getNames() and "HXT" not in res.getNames():
                resname = "ASH"
            elif "OXT" in res.getNames() and res.numAtoms() == 14:
                resname = "ASH"
            elif "H1" in res.getNames() and res.numAtoms() == 15:
                resname = "ASH"
        elif resname == "CL":
            resname = " CL"
        res.setResname(resname)
    return structure


def MoveBackbone(initial_structure, zmatrix, initial_part_of_the_protein, final_part_of_the_protein):
    return
